"""

md_tests.py

A series of tests to validate basic repo functionality and verify either "correct"
inference behavior, or - when operating in environments other than the training
environment - acceptable deviation from the correct results.

This module should not depend on anything else in this repo outside of the 
tests themselves, even if it means some duplicated code (e.g. for downloading files),
since much of what it tries to test is, e.g., imports.

"""

#%% Imports and constants

### Only standard imports belong here, not MD-specific imports ###

import os
import json
import glob
import tempfile
import urllib
import urllib.request
import zipfile
import subprocess
import argparse


#%% Classes

class MDTestOptions:
    """
    Options controlling test behavior.
    """
    
    ## Required ##
    
    #: Force CPU execution
    disable_gpu = False
    
    #: If GPU execution is requested, but a GPU is not available, should we error?
    cpu_execution_is_error = False
    
    #: Skip tests related to video processing
    skip_video_tests = False
    
    #: Skip tests launched via Python functions (as opposed to CLIs)
    skip_python_tests = False
    
    #: Skip CLI tests
    skip_cli_tests = False
    
    #: Force a specific folder for temporary input/output
    scratch_dir = None
    
    #: Where does the test data live?
    test_data_url = 'https://lila.science/public/md-test-package.zip'
    
    #: Download test data even if it appears to have already been downloaded
    force_data_download = False
    
    #: Unzip test data even if it appears to have already been unzipped
    force_data_unzip = False
    
    #: By default, any unexpected behavior is an error; this forces most errors to
    #: be treated as warnings.
    warning_mode = False
    
    #: How much deviation from the expected detection coordinates should we allow before
    #: a disrepancy becomes an error?
    max_coord_error = 0.001
    
    #: How much deviation from the expected confidence values should we allow before
    #: a disrepancy becomes an error?    
    max_conf_error = 0.005
    
    #: Current working directory when running CLI tests
    cli_working_dir = None
    
    #: YOLOv5 installation, only relevant if we're testing run_inference_with_yolov5_val. 
    #:
    #: If this is None, we'll skip that test.
    yolo_working_folder = None

# ...class MDTestOptions()


#%% Support functions

def get_expected_results_filename(gpu_is_available):
    """
    Expected results vary just a little across inference environments, particularly
    between PT 1.x and 2.x, so when making sure things are working acceptably, we 
    compare to a reference file that matches the current environment.
    
    This function gets the correct filename to compare to current results, depending
    on whether a GPU is available.
    
    Args:
        gpu_is_available (bool): whether a GPU is available
        
    Returns:
        str: relative filename of the results file we should use (within the test
        data zipfile)
    """
    
    if gpu_is_available:
        hw_string = 'gpu'
    else:
        hw_string = 'cpu'
    import torch
    torch_version = str(torch.__version__)
    if torch_version.startswith('1'):
        assert torch_version == '1.10.1', 'Only tested against PT 1.10.1 and PT 2.x'
        pt_string = 'pt1.10.1'
    else:
        assert torch_version.startswith('2'), 'Unknown torch version: {}'.format(torch_version)
        pt_string = 'pt2.x'
    
    # A hack for now to account for the fact that even with acceleration enabled and PT2 
    # installed, Apple silicon appears to provide the same results as CPU/PT1 inference
    try:
        import torch
        m1_inference = torch.backends.mps.is_built and torch.backends.mps.is_available()
        if m1_inference:
            hw_string = 'cpu'
        pt_string = 'pt1.10.1'
    except Exception:
        pass
    
    return 'md-test-results-{}-{}.json'.format(hw_string,pt_string)
    
    
def download_test_data(options=None):
    """
    Downloads the test zipfile if necessary, unzips if necessary.
    
    Args:
        options (MDTestOptions, optional): see MDTestOptions for details
        
    Returns:
        MDTestOptions: the same object passed in as input, or the options that
        were used if [options] was supplied as None
    """

    if options is None:
        options = MDTestOptions()
        
    if options.scratch_dir is None:        
        tempdir_base = tempfile.gettempdir()
        scratch_dir = os.path.join(tempdir_base,'md-tests')
    else:
        scratch_dir = options.scratch_dir
    
    os.makedirs(scratch_dir,exist_ok=True)    
    
    # See whether we've already downloaded the data zipfile
    download_zipfile = True        
    if not options.force_data_download:
        local_zipfile = os.path.join(scratch_dir,options.test_data_url.split('/')[-1])
        if os.path.isfile(local_zipfile):
            url_info = urllib.request.urlopen(options.test_data_url).info()
            remote_size = int(url_info['Content-Length'])
            target_file_size = os.path.getsize(local_zipfile)
            if remote_size == target_file_size:
                download_zipfile = False
    
    if download_zipfile:
        print('Downloading test data zipfile')
        urllib.request.urlretrieve(options.test_data_url, local_zipfile)
        print('Finished download to {}'.format(local_zipfile))
    else:
        print('Bypassing test data zipfile download for {}'.format(local_zipfile))
    
    
    ## Unzip data
    
    zipf = zipfile.ZipFile(local_zipfile)    
    zip_contents = zipf.filelist
    
    # file_info = zip_contents[1]
    for file_info in zip_contents:
        
        expected_size = file_info.file_size
        if expected_size == 0:
            continue
        fn_relative = file_info.filename
        target_file = os.path.join(scratch_dir,fn_relative)
        unzip_file = True
        if (not options.force_data_unzip) and os.path.isfile(target_file):
            existing_file_size = os.path.getsize(target_file)
            if existing_file_size == expected_size:
                unzip_file = False
        if unzip_file:
            os.makedirs(os.path.dirname(target_file),exist_ok=True)
            with open(target_file,'wb') as f:
                f.write(zipf.read(fn_relative))
            
    # ...for each file in the zipfile
    
    # Warn if file are present that aren't expected
    test_files = glob.glob(os.path.join(scratch_dir,'**/*'), recursive=True)
    test_files = [os.path.relpath(fn,scratch_dir).replace('\\','/') for fn in test_files]
    test_files_set = set(test_files)
    expected_images_set = set(zipf.namelist())
    for fn in expected_images_set:
        if fn.endswith('/'):
            continue
        assert fn in test_files_set, 'File {} is missing from the test image folder'.format(fn)
    
    # Populate the test options with test data information
    options.scratch_dir = scratch_dir
    options.all_test_files = test_files
    options.test_images = [fn for fn in test_files if os.path.splitext(fn.lower())[1] in ('.jpg','.jpeg','.png')]
    options.test_videos = [fn for fn in test_files if os.path.splitext(fn.lower())[1] in ('.mp4','.avi')]    
    options.test_videos = [fn for fn in options.test_videos if 'rendered' not in fn]
        
    print('Finished unzipping and enumerating test data')
    
    return options

# ...def download_test_data(...)


def is_gpu_available(verbose=True):
    """
    Checks whether a GPU (including M1/M2 MPS) is available.
    
    Args:
        verbose (bool, optional): enable additional debug console output
    
    Returns:
        bool: whether a GPU is available
    """
    
    # Import torch inside this function, so we have a chance to set CUDA_VISIBLE_DEVICES
    # before checking GPU availability.
    import torch
    gpu_available = torch.cuda.is_available()
    
    if gpu_available:
        if verbose:
            print('CUDA available: {}'.format(gpu_available))
            device_ids = list(range(torch.cuda.device_count()))
            if len(device_ids) > 1:
                print('Found multiple devices: {}'.format(str(device_ids)))
    else:
        try:
            gpu_available = torch.backends.mps.is_built and torch.backends.mps.is_available()
        except AttributeError:
            pass
        if gpu_available:
            print('Metal performance shaders available')
    
    if not gpu_available:
        print('No GPU available')
        
    return gpu_available            
        
    
#%% CLI functions

# These are copied from process_utils.py to avoid imports outside of the test
# functions.

os.environ["PYTHONUNBUFFERED"] = "1"

def execute(cmd):
    """
    Runs [cmd] (a single string) in a shell, yielding each line of output to the caller.
    
    Args:
        cmd (str): command to run
    
    Returns:
        int: the command's return code, always zero, otherwise a CalledProcessError is raised
    """
 
    # https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             shell=True, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
    return return_code


def execute_and_print(cmd,print_output=True):
    """
    Runs [cmd] (a single string) in a shell, capturing (and optionally printing) output.
    
    Args:
        cmd (str): command to run
        print_output (bool, optional): whether to print output from [cmd]
    
    Returns:
        dict: a dictionary with fields "status" (the process return code) and "output"
        (the content of stdout)
    """

    to_return = {'status':'unknown','output':''}
    output=[]
    try:
        for s in execute(cmd):
            output.append(s)
            if print_output:
                print(s,end='',flush=True)
        to_return['status'] = 0
    except subprocess.CalledProcessError as cpe:
        print('execute_and_print caught error: {}'.format(cpe.output))
        to_return['status'] = cpe.returncode
    to_return['output'] = output
   
    return to_return


#%% Python tests

def run_python_tests(options):
    """
    Runs Python-based (as opposed to CLI-based) package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    print('\n*** Starting module tests ***\n')
    
    ## Prepare data
    
    download_test_data(options)
    
    
    ## Run inference on an image
    
    from detection import run_detector
    from md_visualization import visualization_utils as vis_utils
    model_file = 'MDV5A'
    image_fn = os.path.join(options.scratch_dir,options.test_images[0])
    model = run_detector.load_detector(model_file)
    pil_im = vis_utils.load_image(image_fn)
    result = model.generate_detections_one_image(pil_im) # noqa

    
    ## Run inference on a folder
    
    from detection.run_detector_batch import load_and_run_detector_batch,write_results_to_file
    from md_utils import path_utils

    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    assert os.path.isdir(image_folder), 'Test image folder {} is not available'.format(image_folder)
    inference_output_file = os.path.join(options.scratch_dir,'folder_inference_output.json')
    image_file_names = path_utils.find_images(image_folder,recursive=True)
    results = load_and_run_detector_batch('MDV5A', image_file_names, quiet=True)
    _ = write_results_to_file(results,inference_output_file,
                              relative_path_base=image_folder,detector_file=model_file)

    # Read results
    with open(inference_output_file,'r') as f:
        results_from_file = json.load(f) # noqa
    

    ## Verify results

    # Read expected results
    expected_results_filename = get_expected_results_filename(is_gpu_available(verbose=False))
    
    with open(os.path.join(options.scratch_dir,expected_results_filename),'r') as f:
        expected_results = json.load(f)
            
    filename_to_results = {im['file'].replace('\\','/'):im for im in results_from_file['images']}
    filename_to_results_expected = {im['file'].replace('\\','/'):im for im in expected_results['images']}
    
    assert len(filename_to_results) == len(filename_to_results_expected), \
        'Error: expected {} files in results, found {}'.format(
            len(filename_to_results_expected),
            len(filename_to_results))
    
    max_coord_error = 0
    max_conf_error = 0
    
    # fn = next(iter(filename_to_results.keys()))
    for fn in filename_to_results.keys():
                
        actual_image_results = filename_to_results[fn]
        expected_image_results = filename_to_results_expected[fn]
        
        if 'failure' in actual_image_results:
            assert 'failure' in expected_image_results and \
                'detections' not in actual_image_results and \
                'detections' not in expected_image_results
            continue
        assert 'failure' not in expected_image_results
        
        actual_detections = actual_image_results['detections']
        expected_detections = expected_image_results['detections']
        
        s = 'expected {} detections for file {}, found {}'.format(
            len(expected_detections),fn,len(actual_detections))
        s += '\nExpected results file: {}\nActual results file: {}'.format(
            expected_results_filename,inference_output_file)
        
        if options.warning_mode:
            if len(actual_detections) != len(expected_detections):
                print('Warning: {}'.format(s))
            continue
        assert len(actual_detections) == len(expected_detections), \
            'Error: {}'.format(s)
        
        # i_det = 0
        for i_det in range(0,len(actual_detections)):
            actual_det = actual_detections[i_det]
            expected_det = expected_detections[i_det]
            assert actual_det['category'] == expected_det['category']
            conf_err = abs(actual_det['conf'] - expected_det['conf'])
            coord_differences = []
            for i_coord in range(0,4):
                coord_differences.append(abs(actual_det['bbox'][i_coord]-expected_det['bbox'][i_coord]))
            coord_err = max(coord_differences)
            
            if conf_err > max_conf_error:
                max_conf_error = conf_err
            if coord_err > max_coord_error:
                max_coord_error = coord_err
        
        # ...for each detection
        
    # ...for each image
    
    if not options.warning_mode:
        
        assert max_conf_error <= options.max_conf_error, \
            'Confidence error {} is greater than allowable ({})'.format(
                max_conf_error,options.max_conf_error)
        
        assert max_coord_error <= options.max_coord_error, \
            'Coord error {} is greater than allowable ({})'.format(
                max_coord_error,options.max_coord_error)
        
    print('Max conf error: {}'.format(max_conf_error))
    print('Max coord error: {}'.format(max_coord_error))


    ## Postprocess results
    
    from api.batch_processing.postprocessing.postprocess_batch_results import \
        PostProcessingOptions,process_batch_results
    postprocessing_options = PostProcessingOptions()
    
    postprocessing_options.api_output_file = inference_output_file
    postprocessing_options.output_dir = os.path.join(options.scratch_dir,'postprocessing_output')
    postprocessing_options.image_base_dir = image_folder
    
    postprocessing_results = process_batch_results(postprocessing_options)
    assert os.path.isfile(postprocessing_results.output_html_file), \
        'Postprocessing output file {} not found'.format(postprocessing_results.output_html_file)
    
    
    ## Partial RDE test
    
    from api.batch_processing.postprocessing.repeat_detection_elimination.repeat_detections_core import \
        RepeatDetectionOptions,find_repeat_detections
    
    rde_options = RepeatDetectionOptions()
    rde_options.occurrenceThreshold = 2
    rde_options.confidenceMin = 0.001
    rde_options.outputBase = os.path.join(options.scratch_dir,'rde_working_dir')
    rde_options.imageBase = image_folder
    rde_output_file = inference_output_file.replace('.json','_filtered.json')
    assert rde_output_file != inference_output_file
    rde_results = find_repeat_detections(inference_output_file, rde_output_file, rde_options)
    assert os.path.isfile(rde_results.filterFile),\
        'Could not find RDE output file {}'.format(rde_results.filterFile)
        
    
    # TODO: add remove_repeat_detections test here
    #
    # It's already tested in the CLI tests, so this is not urgent.
    
    if not options.skip_video_tests:
        
        ## Video test (single video)
       
        from detection.process_video import ProcessVideoOptions, process_video
        
        video_options = ProcessVideoOptions()
        video_options.model_file = 'MDV5A'
        video_options.input_video_file = os.path.join(options.scratch_dir,options.test_videos[0])
        video_options.output_json_file = os.path.join(options.scratch_dir,'single_video_output.json')
        video_options.output_video_file = os.path.join(options.scratch_dir,'video_scratch/rendered_video.mp4')
        video_options.frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder')
        video_options.frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder')    
        video_options.render_output_video = True
        # video_options.keep_rendered_frames = False
        # video_options.keep_rendered_frames = False
        video_options.force_extracted_frame_folder_deletion = True
        video_options.force_rendered_frame_folder_deletion = True
        # video_options.reuse_results_if_available = False
        # video_options.reuse_frames_if_available = False
        video_options.recursive = True
        video_options.verbose = False
        video_options.fourcc = 'mp4v'
        # video_options.rendering_confidence_threshold = None
        # video_options.json_confidence_threshold = 0.005
        video_options.frame_sample = 5    
        video_options.n_cores = 5
        # video_options.debug_max_frames = -1
        # video_options.class_mapping_filename = None
        
        _ = process_video(video_options)
    
        assert os.path.isfile(video_options.output_video_file), \
            'Python video test failed to render output video file'
        assert os.path.isfile(video_options.output_json_file), \
            'Python video test failed to render output .json file'
            
        
        ## Video test (folder)
        
        from detection.process_video import ProcessVideoOptions, process_video_folder
        
        video_options = ProcessVideoOptions()
        video_options.model_file = 'MDV5A'
        video_options.input_video_file = os.path.join(options.scratch_dir,
                                                      os.path.dirname(options.test_videos[0]))
        video_options.output_json_file = os.path.join(options.scratch_dir,'video_folder_output.json')
        # video_options.output_video_file = None
        video_options.frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder')
        video_options.frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder')    
        video_options.render_output_video = False
        # video_options.keep_rendered_frames = False
        # video_options.keep_rendered_frames = False
        video_options.force_extracted_frame_folder_deletion = True
        video_options.force_rendered_frame_folder_deletion = True
        # video_options.reuse_results_if_available = False
        # video_options.reuse_frames_if_available = False
        video_options.recursive = True
        video_options.verbose = False
        # video_options.fourcc = None
        # video_options.rendering_confidence_threshold = None
        # video_options.json_confidence_threshold = 0.005
        video_options.frame_sample = 5    
        video_options.n_cores = 5
        # video_options.debug_max_frames = -1
        # video_options.class_mapping_filename = None
        
        _ = process_video_folder(video_options)
    
        assert os.path.isfile(video_options.output_json_file), \
            'Python video test failed to render output .json file'
        
    # ...if we're not skipping video tests
    
    print('\n*** Finished module tests ***\n')

# ...def run_python_tests(...)


#%% Command-line tests

def run_cli_tests(options):
    """
    Runs CLI (as opposed to Python-based) package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    print('\n*** Starting CLI tests ***\n')
    
    ## chdir if necessary
    
    if options.cli_working_dir is not None:
        os.chdir(options.cli_working_dir)
    
    
    ## Prepare data
    
    download_test_data(options)
    
    
    ## Run inference on an image
    
    model_file = 'MDV5A'
    image_fn = os.path.join(options.scratch_dir,options.test_images[0])
    output_dir = os.path.join(options.scratch_dir,'single_image_test')
    if options.cli_working_dir is None:
        cmd = 'python -m detection.run_detector'
    else:
        cmd = 'python detection/run_detector.py'
    cmd += ' {} --image_file {} --output_dir {}'.format(
        model_file,image_fn,output_dir)
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    if options.cpu_execution_is_error:
        gpu_available_via_cli = False
        for s in cmd_results['output']:
            if 'GPU available: True' in s:
                gpu_available_via_cli = True
                break
        if not gpu_available_via_cli:
            raise Exception('GPU execution is required, but not available')

    
    ## Run inference on a folder
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    assert os.path.isdir(image_folder), 'Test image folder {} is not available'.format(image_folder)
    inference_output_file = os.path.join(options.scratch_dir,'folder_inference_output.json')
    if options.cli_working_dir is None:
        cmd = 'python -m detection.run_detector_batch'
    else:
        cmd = 'python detection/run_detector_batch.py'
    cmd += ' {} {} {} --recursive'.format(
        model_file,image_folder,inference_output_file)
    cmd += ' --output_relative_filenames --quiet --include_image_size'
    cmd += ' --include_image_timestamp --include_exif_data'
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    # Make sure a coherent file got written out, but don't verify the results, leave that
    # to the Python tests.
    with open(inference_output_file,'r') as f:
        results_from_file = json.load(f) # noqa
    

    ## Postprocessing
    
    postprocessing_output_dir = os.path.join(options.scratch_dir,'postprocessing_output_cli')
    
    if options.cli_working_dir is None:
        cmd = 'python -m api.batch_processing.postprocessing.postprocess_batch_results'
    else:
        cmd = 'python api/batch_processing/postprocessing/postprocess_batch_results.py'
    cmd += ' {} {}'.format(
        inference_output_file,postprocessing_output_dir)
    cmd += ' --image_base_dir {}'.format(image_folder)
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
                
    
    ## RDE
    
    rde_output_dir = os.path.join(options.scratch_dir,'rde_output_cli')
    
    if options.cli_working_dir is None:
        cmd = 'python -m api.batch_processing.postprocessing.repeat_detection_elimination.find_repeat_detections'
    else:
        cmd = 'python  api/batch_processing/postprocessing/repeat_detection_elimination/find_repeat_detections.py'
    cmd += ' {}'.format(inference_output_file)
    cmd += ' --imageBase {}'.format(image_folder)
    cmd += ' --outputBase {}'.format(rde_output_dir)
    cmd += ' --occurrenceThreshold 1' # Use an absurd number here to make sure we get some suspicious detections
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)    
    
    # Find the latest filtering folder
    filtering_output_dir = os.listdir(rde_output_dir)
    filtering_output_dir = [fn for fn in filtering_output_dir if fn.startswith('filtering_')]
    filtering_output_dir = [os.path.join(rde_output_dir,fn) for fn in filtering_output_dir]
    filtering_output_dir = [fn for fn in filtering_output_dir if os.path.isdir(fn)]
    filtering_output_dir = sorted(filtering_output_dir)[-1]
    
    print('Using RDE filtering folder {}'.format(filtering_output_dir))
    
    filtered_output_file = inference_output_file.replace('.json','_filtered.json')
    
    if options.cli_working_dir is None:
        cmd = 'python -m api.batch_processing.postprocessing.repeat_detection_elimination.remove_repeat_detections'
    else:
        cmd = 'python  api/batch_processing/postprocessing/repeat_detection_elimination/remove_repeat_detections.py'
    cmd += ' {} {} {}'.format(inference_output_file,filtered_output_file,filtering_output_dir)
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    assert os.path.isfile(filtered_output_file), \
        'Could not find RDE output file {}'.format(filtered_output_file)
    
    
    ## Run inference on a folder (tiled)
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    tiling_folder = os.path.join(options.scratch_dir,'tiling-folder')
    inference_output_file_tiled = os.path.join(options.scratch_dir,'folder_inference_output_tiled.json')
    if options.cli_working_dir is None:
        cmd = 'python -m detection.run_tiled_inference'
    else:
        cmd = 'python detection/run_tiled_inference.py'
    cmd += ' {} {} {} {}'.format(
        model_file,image_folder,tiling_folder,inference_output_file_tiled)
    cmd += ' --overwrite_handling overwrite'
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    with open(inference_output_file_tiled,'r') as f:
        results_from_file = json.load(f) # noqa
        
    
    ## Run inference on a folder (augmented)
    
    if options.yolo_working_folder is None:
        
        print('Bypassing YOLOv5 val tests, no yolo folder supplied')
        
    else:
    
        image_folder = os.path.join(options.scratch_dir,'md-test-images')
        yolo_results_folder = os.path.join(options.scratch_dir,'yolo-output-folder')
        yolo_symlink_folder = os.path.join(options.scratch_dir,'yolo-symlink_folder')
        inference_output_file_yolo_val = os.path.join(options.scratch_dir,'folder_inference_output_yolo_val.json')
        if options.cli_working_dir is None:
            cmd = 'python -m detection.run_inference_with_yolov5_val'
        else:
            cmd = 'python detection/run_inference_with_yolov5_val.py'
        cmd += ' {} {} {}'.format(
            model_file,image_folder,inference_output_file_yolo_val)
        cmd += ' --yolo_working_folder {}'.format(options.yolo_working_folder)
        cmd += ' --yolo_results_folder {}'.format(yolo_results_folder)
        cmd += ' --symlink_folder {}'.format(yolo_symlink_folder)
        cmd += ' --augment_enabled 1'
        # cmd += ' --no_use_symlinks'
        cmd += ' --overwrite_handling overwrite'
        print('Running: {}'.format(cmd))
        cmd_results = execute_and_print(cmd)
        
        with open(inference_output_file_yolo_val,'r') as f:
            results_from_file = json.load(f) # noqa
        
        
    if not options.skip_video_tests:
            
        ## Video test
        
        model_file = 'MDV5A'
        video_inference_output_file = os.path.join(options.scratch_dir,'video_inference_output.json')
        output_video_file = os.path.join(options.scratch_dir,'video_scratch/cli_rendered_video.mp4')
        frame_folder = os.path.join(options.scratch_dir,'video_scratch/frame_folder_cli')
        frame_rendering_folder = os.path.join(options.scratch_dir,'video_scratch/rendered_frame_folder_cli')        
        
        video_fn = os.path.join(options.scratch_dir,options.test_videos[-1])
        output_dir = os.path.join(options.scratch_dir,'single_video_test_cli')
        if options.cli_working_dir is None:
            cmd = 'python -m detection.process_video'
        else:
            cmd = 'python detection/process_video.py'
        cmd += ' {} {}'.format(model_file,video_fn)
        cmd += ' --frame_folder {} --frame_rendering_folder {} --output_json_file {} --output_video_file {}'.format(
            frame_folder,frame_rendering_folder,video_inference_output_file,output_video_file)
        cmd += ' --render_output_video --fourcc mp4v'
        cmd += ' --force_extracted_frame_folder_deletion --force_rendered_frame_folder_deletion --n_cores 5 --frame_sample 3'
        print('Running: {}'.format(cmd))
        cmd_results = execute_and_print(cmd)

    # ...if we're not skipping video tests
    
    
    ## Run inference on a folder (again, so we can do a comparison)
    
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    model_file = 'MDV5B'
    inference_output_file_alt = os.path.join(options.scratch_dir,'folder_inference_output_alt.json')
    if options.cli_working_dir is None:
        cmd = 'python -m detection.run_detector_batch'
    else:
        cmd = 'python detection/run_detector_batch.py'
    cmd += ' {} {} {} --recursive'.format(
        model_file,image_folder,inference_output_file_alt)
    cmd += ' --output_relative_filenames --quiet --include_image_size'
    cmd += ' --include_image_timestamp --include_exif_data'
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    with open(inference_output_file_alt,'r') as f:
        results_from_file = json.load(f) # noqa
    
    
    ## Compare the two files
    
    comparison_output_folder = os.path.join(options.scratch_dir,'results_comparison')
    image_folder = os.path.join(options.scratch_dir,'md-test-images')
    results_files_string = '"{}" "{}"'.format(
        inference_output_file,inference_output_file_alt)
    if options.cli_working_dir is None:
        cmd = 'python -m api.batch_processing.postprocessing.compare_batch_results'
    else:
        cmd = 'python api/batch_processing/postprocessing/compare_batch_results.py'
    cmd += ' {} {} {}'.format(comparison_output_folder,image_folder,results_files_string)
    print('Running: {}'.format(cmd))
    cmd_results = execute_and_print(cmd)
    
    assert cmd_results['status'] == 0, 'Error generating comparison HTML'
    assert os.path.isfile(os.path.join(comparison_output_folder,'index.html')), \
        'Failed to generate comparison HTML'
    
    print('\n*** Finished CLI tests ***\n')
    
# ...def run_cli_tests(...)


#%% Main test wrapper

def run_tests(options):
    """
    Runs Python-based and/or CLI-based package tests.
    
    Args:
        options (MDTestOptions): see MDTestOptions for details
    """
    
    # Prepare data folder
    download_test_data(options)    
    
    if options.disable_gpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
        
    # Verify GPU
    gpu_available = is_gpu_available()
    
    # If the GPU is required and isn't available, error
    if options.cpu_execution_is_error and (not gpu_available):
        raise ValueError('GPU not available, and cpu_execution_is_error is set')
    
    # If the GPU should be disabled, verify that it is
    if options.disable_gpu:
        assert (not gpu_available), 'CPU execution specified, but the GPU appears to be available'
        
    # Run python tests
    if not options.skip_python_tests:
        run_python_tests(options)
    
    # Run CLI tests
    if not options.skip_cli_tests:
        run_cli_tests(options)


#%% Interactive driver

if False:
    
    pass

    #%%
    
    options = MDTestOptions()
    
    options.disable_gpu = False
    options.cpu_execution_is_error = False
    options.skip_video_tests = False
    options.skip_python_tests = False
    options.skip_cli_tests = False
    options.scratch_dir = None
    options.test_data_url = 'https://lila.science/public/md-test-package.zip'
    options.force_data_download = False
    options.force_data_unzip = False
    options.warning_mode = True
    options.max_coord_error = 0.001
    options.max_conf_error = 0.005
    options.cli_working_dir = r'c:\git\MegaDetector'
    options.yolo_working_folder = r'c:\git\yolov5'


    #%%
    
    run_tests(options)
    
    
#%% Command-line driver

def main():

    options = MDTestOptions()
    
    parser = argparse.ArgumentParser(
        description='MegaDetector test suite')
    
    parser.add_argument(
        '--disable_gpu',
        action='store_true',
        help='Disable GPU operation')
    
    parser.add_argument(
        '--cpu_execution_is_error',
        action='store_true',
        help='Fail if the GPU appears not to be available')
    
    parser.add_argument(
        '--scratch_dir',
        default=None,
        type=str,
        help='Directory for temporary storage (defaults to system temp dir)')
    
    parser.add_argument(
        '--skip_video_tests',
        action='store_true',
        help='Skip tests related to video (which can be slow)')
        
    parser.add_argument(
        '--skip_python_tests',
        action='store_true',
        help='Skip python tests')
        
    parser.add_argument(
        '--skip_cli_tests',
        action='store_true',
        help='Skip CLI tests')
        
    parser.add_argument(
        '--force_data_download',
        action='store_true',
        help='Force download of the test data file, even if it\'s already available')
    
    parser.add_argument(
        '--force_data_unzip',
        action='store_true',
        help='Force extraction of all files in the test data file, even if they\'re already available')
    
    parser.add_argument(
        '--warning_mode',
        action='store_true',
        help='Turns numeric/content errors into warnings')
    
    parser.add_argument(
        '--max_conf_error',
        type=float,
        default=options.max_conf_error,
        help='Maximum tolerable confidence value deviation from expected (default {})'.format(
            options.max_conf_error))
    
    parser.add_argument(
        '--max_coord_error',
        type=float,
        default=options.max_coord_error,
        help='Maximum tolerable coordinate value deviation from expected (default {})'.format(
            options.max_coord_error))

    parser.add_argument(
        '--cli_working_dir',
        type=str,
        default=None,
        help='Working directory for CLI tests')

    # token used for linting
    #
    # no_arguments_required
        
    args = parser.parse_args()
        
    options.disable_gpu = args.disable_gpu
    options.cpu_execution_is_error = args.cpu_execution_is_error
    options.skip_video_tests = args.skip_video_tests
    options.skip_python_tests = args.skip_python_tests
    options.skip_cli_tests = args.skip_cli_tests
    options.scratch_dir = args.scratch_dir
    options.warning_mode = args.warning_mode
    options.force_data_download = args.force_data_download
    options.max_conf_error = args.max_conf_error
    options.max_coord_error = args.max_coord_error
    options.cli_working_dir = args.cli_working_dir

    run_tests(options)
    
if __name__ == '__main__':
    main()



"""

simple_image_download.py

Web image downloader, used in preview_lila_taxonomy.py

Slightly modified from:

https://github.com/RiddlerQ/simple_image_download

"""

#%% Imports

import os
import urllib
import requests
import magic
from urllib.parse import quote
import random


#%% Constants

BASE_URL = 'https://www.google.com/search?q='
GOOGLE_PICTURE_ID = '''&biw=1536&bih=674&tbm=isch&sxsrf=ACYBGNSXXpS6YmAKUiLKKBs6xWb4uUY5gA:1581168823770&source=lnms&sa=X&ved=0ahUKEwioj8jwiMLnAhW9AhAIHbXTBMMQ_AUI3QUoAQ'''
HEADERS = {
    'User-Agent':
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.90 Safari/537.36"
}
SCANNER_COUNTER = None


#%% Support functions

def generate_search_url(keywords):
    keywords_to_search = [str(item).strip() for item in keywords.split(',')][0].split()
    keywords_count = len(keywords_to_search)
    return keywords_to_search, keywords_count


def generate_urls(search):
    """
    Generate Google search URLs for all tokens in the list [search]
    """
    
    return [(BASE_URL+quote(word)+GOOGLE_PICTURE_ID) for word in search]


def check_webpage(url):
    checked_url = None
    try:
        request = requests.get(url, allow_redirects=True, timeout=10)
        if 'html' not in str(request.content):
            checked_url = request
    except Exception as err:
        print(err)        
    return checked_url


def scan_webpage(webpage, extensions, timer):
    """
    Scan for pictures to download based on keywords
    """
    
    global SCANNER_COUNTER
    scanner = webpage.find
    found = False
    counter = 0
    while counter < timer:
        new_line = scanner('"https://', SCANNER_COUNTER + 1)  # How Many New lines
        SCANNER_COUNTER = scanner('"', new_line + 1)  # Ends of line
        buffer = scanner('\\', new_line + 1, SCANNER_COUNTER)
        if buffer != -1:
            object_raw = webpage[new_line + 1:buffer]
        else:
            object_raw = webpage[new_line + 1:SCANNER_COUNTER]
        if any(extension in object_raw for extension in extensions):
            found = True
            break
        counter += 1
    if found:
        object_ready = check_webpage(object_raw)
        return object_ready


#%% Main class

class Downloader:
    """
    Main Downloader
    ::param extension:iterable of Files extensions
    """
    def __init__(self, extensions=None):
        if extensions:
            self._extensions = set(*[extensions])
        else:
            self._extensions = {'.jpg', '.png', '.ico', '.gif', '.jpeg'}
        self._directory = "simple_images/"
        self.get_dirs = set()
        self._cached_urls = {}

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value

    @property
    def cached_urls(self):
        return self._cached_urls

    @property
    def extensions(self):
        return self._extensions

    @extensions.setter
    def extensions(self, value):
        self._extensions = set([value])


    def get_urls(self):
        return [self._cached_urls[url][1].url
                for url in self._cached_urls]

    def _download_page(self, url):
        req = urllib.request.Request(url, headers=HEADERS)
        resp = urllib.request.urlopen(req)
        resp_data = str(resp.read())
        return resp_data

    def search_urls(self, keywords, limit=1, verbose=False, cache=True, timer=None):
        cache_out = {}
        search, count = generate_search_url(keywords)
        urls_ = generate_urls(search)
        timer = timer if timer else 1000
        # max_progressbar = count * (list(range(limit+1))[-1]+1)
        
        # bar = progressbar.ProgressBar(maxval=max_progressbar,
        #                               widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage()]).start()
        i = 0
        while i < count:
            global SCANNER_COUNTER
            SCANNER_COUNTER = -1
            url = urls_[i]
            path = self.generate_dir(search[i])
            raw_html = self._download_page(url) # Download the entire page from the google Picture search
            for _ in range(limit+1):
                webpage_url = scan_webpage(raw_html, self._extensions, timer)
                if webpage_url:
                    file_name = Downloader.gen_fn(webpage_url, search[i])
                    cache_out[file_name] = [path, webpage_url]
                else:
                    pass
                # bar.update(bar.currval + 1)
            i += 1
        # bar.finish()
        if verbose:
            for url in cache_out:
                print(url)
        if cache:
            self._cached_urls = cache_out
        if not cache_out:
            print('==='*15 + ' < ' + 'NO PICTURES FOUND' + ' > ' + '==='*15)
        return cache_out

    def download(self, keywords=None, limit=1, verbose=False, cache=True, download_cache=False, 
                 timer=None):
        if not download_cache:
            content = self.search_urls(keywords, limit, verbose, cache, timer)
        else:
            content = self._cached_urls
            if not content:
                print('Downloader has not URLs saved in Memory yet, run Downloader.search_urls to find pics first')
        paths = []                
        for name, (path, url) in content.items():
            fullpath = os.path.join(path, name)
            paths.append(fullpath)
            with open(fullpath, 'wb') as file:
                file.write(url.content)
            if verbose:
                print(f'File Name={name}, Downloaded from {url.url}')            
        return paths
    
    def _create_directories(self, name):
        dir_path = os.path.join(self._directory, name)
        try:
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)
        except OSError:
            raise
        self.get_dirs.update([name])
        return

    def generate_dir(self, dir_name):
        """Generate Path and Directory, also check if Directory exists or not """
        dir_name = dir_name.replace(" ", "_")
        if dir_name in self.get_dirs:
            pass
        else:
            self._create_directories(dir_name)
        return os.path.join(self._directory,dir_name)

    @staticmethod
    def gen_fn(check, name):
        """Create a file name string and generate a random identifiers otherwise won't import same pic twice"""
        id = str(hex(random.randrange(1000)))
        mime = magic.Magic(mime=True)
        file_type = mime.from_buffer(check.content)
        file_extension = f'.{file_type.split("/")[1]}'
        file_name = str(name) + "_" + id[2:] + file_extension
        return file_name

    def flush_cache(self):
        """Clear the Downloader instance cache"""
        self._cached_urls = set()



"""

coco_to_labelme.py

Converts a COCO dataset to labelme format (one .json per image file).

If you want to convert YOLO-formatted data to labelme format, use yolo_to_coco, then 
coco_to_labelme.

"""

#%% Imports and constants

import os
import json

from tqdm import tqdm
from collections import defaultdict

from md_visualization.visualization_utils import open_image


#%% Functions

def get_labelme_dict_for_image_from_coco_record(im,annotations,categories,info=None):
    """
    For the given image struct in COCO format and associated list of annotations, reformats the 
    detections into labelme format.  
    
    Args:
        im (dict): image dict, as loaded from a COCO .json file; 'height' and 'width' are required
        annotations (list): a list of annotations that refer to this image (this function errors if 
            that's not the case)
        categories (list): a list of category in dicts in COCO format ({'id':x,'name':'s'})
        info (dict, optional): a dict to store in a non-standard "custom_info"  field in the output
        
    Returns:
        dict: a dict in labelme format, suitable for writing to a labelme .json file
    """
    
    image_base_name = os.path.basename(im['file_name'])
    
    output_dict = {}
    if info is not None:
        output_dict['custom_info'] = info
    output_dict['version'] = '5.3.0a0'
    output_dict['flags'] = {}
    output_dict['shapes'] = []
    output_dict['imagePath'] = image_base_name
    output_dict['imageHeight'] = im['height']
    output_dict['imageWidth'] = im['width']
    output_dict['imageData'] = None
    
    # Store COCO categories in case we want to reconstruct the original IDs later
    output_dict['coco_categories'] = categories
    
    category_id_to_name = {c['id']:c['name'] for c in categories}
    
    if 'flags' in im:
        output_dict['flags'] = im['flags']
        
    # ann = annotations[0]
    for ann in annotations:
        
        assert ann['image_id'] == im['id'], 'Annotation {} does not refer to image {}'.format(
            ann['id'],im['id'])
        
        if 'bbox' not in ann:
            continue
        
        shape = {}
        shape['label'] = category_id_to_name[ann['category_id']] 
        shape['shape_type'] = 'rectangle'
        shape['description'] = ''
        shape['group_id'] = None
        
        # COCO boxes are [x_min, y_min, width_of_box, height_of_box] (absolute)
        # 
        # labelme boxes are [[x0,y0],[x1,y1]] (absolute)
        x0 = ann['bbox'][0]
        y0 = ann['bbox'][1]
        x1 = ann['bbox'][0] + ann['bbox'][2]
        y1 = ann['bbox'][1] + ann['bbox'][3]
        
        shape['points'] = [[x0,y0],[x1,y1]]
        output_dict['shapes'].append(shape)
    
    # ...for each detection
    
    return output_dict

# ...def get_labelme_dict_for_image()


def coco_to_labelme(coco_data,image_base,overwrite=False,bypass_image_size_check=False,verbose=False):
    """
    For all the images in [coco_data] (a dict or a filename), write a .json file in 
    labelme format alongside the corresponding relative path within image_base.    
    """
    
    # Load COCO data if necessary
    if isinstance(coco_data,str):
        with open(coco_data,'r') as f:
            coco_data = json.load(f)
    assert isinstance(coco_data,dict)
     
    
    ## Read image sizes if necessary
    
    if bypass_image_size_check:
        
        print('Bypassing size check')
        
    else:
    
        # TODO: parallelize this loop
        
        print('Reading/validating image sizes...')
        
        # im = coco_data['images'][0]
        for im in tqdm(coco_data['images']):
            
            # Make sure this file exists
            im_full_path = os.path.join(image_base,im['file_name'])
            assert os.path.isfile(im_full_path), 'Image file {} does not exist'.format(im_full_path)
            
            # Load w/h information if necessary
            if 'height' not in im or 'width' not in im:
                
                try:
                    pil_im = open_image(im_full_path)
                    im['width'] = pil_im.width
                    im['height'] = pil_im.height
                except Exception:
                    print('Warning: cannot open image {}'.format(im_full_path))
                    if 'failure' not in im:
                        im['failure'] = 'Failure image access'
    
            # ...if we need to read w/h information
            
        # ...for each image
    
    # ...if we need to load image sizes
    
    
    ## Generate labelme files
    
    print('Generating .json files...')
    
    image_id_to_annotations = defaultdict(list)
    for ann in coco_data['annotations']:
        image_id_to_annotations[ann['image_id']].append(ann)
        
    n_json_files_written = 0
    n_json_files_error = 0
    n_json_files_exist = 0
    
    # Write output
    for im in tqdm(coco_data['images']):
        
        # Skip this image if it failed to load in whatever system generated this COCO file
        skip_image = False
        
        # Errors are represented differently depending on the source
        for error_string in ('failure','error'):
            if (error_string in im) and (im[error_string] is not None):
                if verbose:
                    print('Warning: skipping labelme file generation for failed image {}'.format(
                        im['file_name']))
                skip_image = True
                n_json_files_error += 1
                break
        if skip_image:
            continue
                    
        im_full_path = os.path.join(image_base,im['file_name'])
        json_path = os.path.splitext(im_full_path)[0] + '.json'
        
        if (not overwrite) and (os.path.isfile(json_path)):
            if verbose:
                print('Skipping existing file {}'.format(json_path))
            n_json_files_exist += 1
            continue
    
        annotations_this_image = image_id_to_annotations[im['id']]
        output_dict = get_labelme_dict_for_image_from_coco_record(im,
                                                                  annotations_this_image,
                                                                  coco_data['categories'],
                                                                  info=None)
        
        n_json_files_written += 1
        with open(json_path,'w') as f:
            json.dump(output_dict,f,indent=1)
            
    # ...for each image
    
    print('\nWrote {} .json files (skipped {} for errors, {} because they exist)'.format(
        n_json_files_written,n_json_files_error,n_json_files_exist))
    
# ...def coco_to_labelme()


#%% Interactive driver

if False:
    
    pass

    #%% Configure options
    
    coco_file = \
        r'C:\\temp\\snapshot-exploration\\images\\training-images-good\\training-images-good_from_yolo.json'
    image_folder = os.path.dirname(coco_file)    
    overwrite = True    
    
    
    #%% Programmatic execution
    
    coco_to_labelme(coco_data=coco_file,image_base=image_folder,overwrite=overwrite)

    
    #%% Command-line execution
    
    s = 'python coco_to_labelme.py "{}" "{}"'.format(coco_file,image_folder)
    if overwrite:
        s += ' --overwrite'
        
    print(s)
    import clipboard; clipboard.copy(s)


    #%% Opening labelme
    
    s = 'python labelme {}'.format(image_folder)
    print(s)
    import clipboard; clipboard.copy(s)
    

#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='Convert a COCO database to labelme annotation format')
    
    parser.add_argument(
        'coco_file',
        type=str,
        help='Path to COCO data file (.json)')
    
    parser.add_argument(
        'image_base',
        type=str,
        help='Path to images (also the output folder)')
    
    parser.add_argument(
        '--overwrite',
        action='store_true',
        help='Overwrite existing labelme .json files')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    coco_to_labelme(coco_data=args.coco_file,image_base=args.image_base,overwrite=args.overwrite)    

if __name__ == '__main__':
    main()



"""

 save_the_elephants_survey_A.py

 Convert the .csv file provided for the Save the Elephants Survey A data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

import pandas as pd
import os
import json
import uuid
import time
import humanfriendly
import numpy as np
from tqdm import tqdm

from md_utils.path_utils import find_images

input_base = r'z:/ste_2019_08_drop'
input_metadata_file = os.path.join(input_base,'SURVEY_A.xlsx')

output_base = r'f:/save_the_elephants/survey_a'
output_json_file = os.path.join(output_base,'ste_survey_a.json')
image_directory = os.path.join(input_base,'SURVEY A with False Triggers')
                        
os.makedirs(output_base,exist_ok=True)
assert(os.path.isdir(image_directory))
assert(os.path.isfile(input_metadata_file))

# Handle all unstructured fields in the source data as extra fields in the annotations
mapped_fields = {'No. of Animals in Photo':'num_animals',
                 'No. of new indiviauls (first sighting of new individual)':'num_new_individuals',
                 'Number Adult Males (first sighting of new individual)':'num_adult_males',
                 'Number Adult Females (first sighting of new individual)':'num_adult_females',
                 'Number Adult Unknown (first sighting of new individual)':'num_adult_unknown',
                 'Number Sub-adult Males (first sighting of new individual)':'num_subadult_males',
                 'Number Sub-adult Females (first sighting of new individual)':'num_subadult_females',
                 'Number Sub-adult Unknown (first sighting of new individual)':'num_subadult_unknown',
                 'Number Juvenile (first sighting of new individual)':'num_juvenile',
                 'Number Newborn (first sighting of new individual)':'num_newborn',
                 'Activity':'activity',
                 'Animal ID':'animal_id',
                 'Specific Notes':'notes'}

# photo_type really should be an image property, but there are a few conflicts
# that forced me to handle it as an annotation proprerty
mapped_fields['Photo Type '] = 'photo_type'


#%% Read source data

input_metadata = pd.read_excel(input_metadata_file, sheet_name='9. CT Image')
input_metadata = input_metadata.iloc[2:]

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))


#%% Map filenames to rows, verify image existence

start_time = time.time()

# Maps relative paths to row indices in input_metadata
filenames_to_rows = {}
filenames_with_multiple_annotations = []
missing_images = []

# Build up a map from filenames to a list of rows, checking image existence as we go
for i_row, fn in tqdm(enumerate(input_metadata['Image Name']),total=len(input_metadata)):
    
    # Ignore directories
    if not fn.endswith('.JPG'):
        continue
    
    if fn in filenames_to_rows:
        filenames_with_multiple_annotations.append(fn)
        filenames_to_rows[fn].append(i_row)
    else:
        filenames_to_rows[fn] = [i_row]
        image_path = os.path.join(image_directory, fn)
        if not os.path.isfile(image_path):
            missing_images.append(image_path)
        
elapsed = time.time() - start_time

print('Finished verifying image existence for {} files in {}, found {} filenames with multiple labels, {} missing images'.format(
      len(filenames_to_rows), humanfriendly.format_timespan(elapsed), 
      len(filenames_with_multiple_annotations),len(missing_images)))

        
#%% Make sure the multiple-annotation cases make sense

if False:
    
    #%% 
    
    fn = filenames_with_multiple_annotations[1000] 
    rows = filenames_to_rows[fn]
    assert(len(rows) > 1)
    for i_row in rows:
        print(input_metadata.iloc[i_row]['Species'])
    

#%% Check for images that aren't included in the metadata file

# Enumerate all images
image_full_paths = find_images(image_directory, bRecursive=True)

unannotated_images = []

for iImage, image_path in tqdm(enumerate(image_full_paths),total=len(image_full_paths)):
    relative_path = os.path.relpath(image_path,image_directory)
    if relative_path not in filenames_to_rows:
        unannotated_images.append(relative_path)

print('Finished checking {} images to make sure they\'re in the metadata, found {} unannotated images'.format(
        len(image_full_paths),len(unannotated_images)))


#%% Create CCT dictionaries

images = []
annotations = []
categories = []

image_ids_to_images = {}

category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category
categories.append(empty_category)
next_category_id = 1

start_time = time.time()

# i_image = 0; image_name = list(filenames_to_rows.keys())[i_image]
for image_name in tqdm(list(filenames_to_rows.keys())):

    # Example filename:
    #        
    # 'Site 1_Oloisukut_1\Oloisukut_A11_UP\Service_2\100EK113\EK001382.JPG'
    # 'Site 1_Oloisukut_1\Oloisukut_A11_UP\Service_2.1\100EK113\EK001382.JPG'
    img_id = image_name.replace('\\','/').replace('/','_').replace(' ','_')
    
    row_indices = filenames_to_rows[image_name]
    
    # i_row = row_indices[0]
    for i_row in row_indices:
        
        row = input_metadata.iloc[i_row]
        assert(row['Image Name'] == image_name)
        
        timestamp = row['Date'].strftime("%d/%m/%Y")
        station_label = row['Camera Trap Station Label']
        photo_type = row['Photo Type ']
        if isinstance(photo_type,float):
            photo_type = ''
        photo_type = photo_type.strip().lower()
            
        if img_id in image_ids_to_images:
            
            im = image_ids_to_images[img_id]
            assert im['file_name'] == image_name
            assert im['station_label'] == station_label
            
            # There are a small handful of datetime mismatches across annotations
            # for the same image
            # assert im['datetime'] == timestamp
            if im['datetime'] != timestamp:
                print('Warning: timestamp conflict for image {}: {},{}'.format(
                    image_name,im['datetime'],timestamp))
                
        else:
            
            im = {}
            im['id'] = img_id
            im['file_name'] = image_name
            im['datetime'] = timestamp
            im['station_label'] = station_label
            im['photo_type'] = photo_type
            
            image_ids_to_images[img_id] = im
            images.append(im)
    
        species = row['Species']
        
        if (isinstance(species,float) or \
            (isinstance(species,str) and (len(species) == 0))):
            category_name = 'empty'
        else:
            category_name = species
        
        # Special cases based on the 'photo type' field
        if 'vehicle' in photo_type:
            category_name = 'vehicle'
        # Various spellings of 'community'
        elif 'comm' in photo_type:
            category_name = 'human'
        elif 'camera' in photo_type or 'researcher' in photo_type:
            category_name = 'human'
        elif 'livestock' in photo_type:
            category_name = 'livestock'
        elif 'blank' in photo_type:
            category_name = 'empty'
        elif 'plant movement' in photo_type:
            category_name = 'empty'
            
        category_name = category_name.strip().lower()
            
        # Have we seen this category before?
        if category_name in category_name_to_category:
            category_id = category_name_to_category[category_name]['id'] 
        else:
            category_id = next_category_id
            category = {}
            category['id'] = category_id
            category['name'] = category_name
            category_name_to_category[category_name] = category
            categories.append(category)
            next_category_id += 1
        
        # Create an annotation
        ann = {}        
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']    
        ann['category_id'] = category_id
        
        # fieldname = list(mapped_fields.keys())[0]
        for fieldname in mapped_fields:
            target_field = mapped_fields[fieldname]
            val = row[fieldname]
            if isinstance(val,float) and np.isnan(val):
                val = ''
            else:
                val = str(val).strip()
            ann[target_field] = val
            
        annotations.append(ann)
        
    # ...for each row
                
# ...for each image
    
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'Save the Elephants Survey A'
info['contributor'] = 'Save the Elephants'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate output

from data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
    
sortedCategories, data = integrity_check_json_db.integrity_check_json_db(output_json_file,options)


#%% Preview labels

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 1000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file,image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                        output_dir=os.path.join(output_base,'preview'),
                                                        image_base_dir=image_directory,
                                                        options=viz_options)
os.startfile(html_output_file)


#%% Scrap

if False:

    pass
    
    #%% Find unique photo types
    
    annotations = image_db['annotations']
    photo_types = set()
    for ann in tqdm(annotations):
        photo_types.add(ann['photo_type'])



"""

train_classifier_tf.py

Train an EfficientNet classifier.

Currently the implementation of multi-label multi-class classification is
non-functional.

During training, start tensorboard from within the classification/ directory:
   tensorboard --logdir run --bind_all --samples_per_plugin scalars=0,images=0

"""

#%% Imports and constants

from __future__ import annotations

import argparse
from collections import defaultdict
from collections.abc import Callable, Mapping, MutableMapping, Sequence
from datetime import datetime
import json
import os
from typing import Any, Optional
import uuid

import numpy as np
import sklearn.metrics
import tensorflow as tf
from tensorboard.plugins.hparams import api as hp
import tqdm

from classification.train_utils import (
    HeapItem, recall_from_confusion_matrix, add_to_heap, fig_to_img,
    imgs_with_confidences, load_dataset_csv, prefix_all_keys)
from md_visualization import plot_utils

AUTOTUNE = tf.data.experimental.AUTOTUNE

# match pytorch EfficientNet model names
EFFICIENTNET_MODELS: Mapping[str, Mapping[str, Any]] = {
    'efficientnet-b0': dict(cls='EfficientNetB0', img_size=224, dropout=0.2),
    'efficientnet-b1': dict(cls='EfficientNetB1', img_size=240, dropout=0.2),
    'efficientnet-b2': dict(cls='EfficientNetB2', img_size=260, dropout=0.3),
    'efficientnet-b3': dict(cls='EfficientNetB3', img_size=300, dropout=0.3),
    'efficientnet-b4': dict(cls='EfficientNetB4', img_size=380, dropout=0.4),
    'efficientnet-b5': dict(cls='EfficientNetB5', img_size=456, dropout=0.4),
    'efficientnet-b6': dict(cls='EfficientNetB6', img_size=528, dropout=0.5),
    'efficientnet-b7': dict(cls='EfficientNetB7', img_size=600, dropout=0.5)
}


#%% Example usage

"""    
    python train_classifier_tf.py run_idfg /ssd/crops_sq \
        -m "efficientnet-b0" --pretrained --finetune --label-weighted \
        --epochs 50 --batch-size 512 --lr 1e-4 \
        --seed 123 \
        --logdir run_idfg
"""


#%% Support functions

def create_dataset(
        img_files: Sequence[str],
        labels: Sequence[Any],
        sample_weights: Optional[Sequence[float]] = None,
        img_base_dir: str = '',
        transform: Optional[Callable[[tf.Tensor], Any]] = None,
        target_transform: Optional[Callable[[Any], Any]] = None,
        cache: bool | str = False
        ) -> tf.data.Dataset:
    """
    Create a tf.data.Dataset.

    The dataset returns elements (img, label, img_file, sample_weight) if
    sample_weights is not None, or (img, label, img_file) if
    sample_weights=None.
        img: tf.Tensor, shape [H, W, 3], type uint8
        label: tf.Tensor
        img_file: tf.Tensor, scalar, type str
        sample_weight: tf.Tensor, scalar, type float32

    Args:
        img_files: list of str, relative paths from img_base_dir
        labels: list of int if multilabel=False
        sample_weights: optional list of float
        img_base_dir: str, base directory for images
        transform: optional transform to apply to a single uint8 JPEG image
        target_transform: optional transform to apply to a single label
        cache: bool or str, cache images in memory if True, cache images to
            a file on disk if a str

    Returns: tf.data.Dataset
    """
    
    # images dataset
    img_ds = tf.data.Dataset.from_tensor_slices(img_files)
    img_ds = img_ds.map(lambda p: tf.io.read_file(img_base_dir + os.sep + p),
                        num_parallel_calls=AUTOTUNE)

    # for smaller disk / memory usage, we cache the raw JPEG bytes instead
    # of the decoded Tensor
    if isinstance(cache, str):
        img_ds = img_ds.cache(cache)
    elif cache:
        img_ds = img_ds.cache()

    # convert JPEG bytes to a 3D uint8 Tensor
    # keras EfficientNet already includes normalization from [0, 255] to [0, 1],
    #   so we don't need to do that here
    img_ds = img_ds.map(lambda img: tf.io.decode_jpeg(img, channels=3))

    if transform:
        img_ds = img_ds.map(transform, num_parallel_calls=AUTOTUNE)

    # labels dataset
    labels_ds = tf.data.Dataset.from_tensor_slices(labels)
    if target_transform:
        labels_ds = labels_ds.map(target_transform, num_parallel_calls=AUTOTUNE)

    # img_files dataset
    img_files_ds = tf.data.Dataset.from_tensor_slices(img_files)

    if sample_weights is None:
        return tf.data.Dataset.zip((img_ds, labels_ds, img_files_ds))

    # weights dataset
    weights_ds = tf.data.Dataset.from_tensor_slices(sample_weights)
    return tf.data.Dataset.zip((img_ds, labels_ds, img_files_ds, weights_ds))


def create_dataloaders(
        dataset_csv_path: str,
        label_index_json_path: str,
        splits_json_path: str,
        cropped_images_dir: str,
        img_size: int,
        multilabel: bool,
        label_weighted: bool,
        weight_by_detection_conf: bool | str,
        batch_size: int,
        augment_train: bool,
        cache_splits: Sequence[str]
        ) -> tuple[dict[str, tf.data.Dataset], list[str]]:
    """
    Args:
        dataset_csv_path: str, path to CSV file with columns
            ['dataset', 'location', 'label'], where label is a comma-delimited
            list of labels
        splits_json_path: str, path to JSON file
        augment_train: bool, whether to shuffle/augment the training set
        cache_splits: list of str, splits to cache
            training set is cached at /mnt/tempds/random_file_name
            validation and test sets are cached in memory

    Returns:
        datasets: dict, maps split to DataLoader
        label_names: list of str, label names in order of label id
    """
    
    df, label_names, split_to_locs = load_dataset_csv(
        dataset_csv_path, label_index_json_path, splits_json_path,
        multilabel=multilabel, label_weighted=label_weighted,
        weight_by_detection_conf=weight_by_detection_conf)

    # define the transforms

    # efficientnet data preprocessing:
    # - train:
    #   1) random crop: aspect_ratio_range=(0.75, 1.33), area_range=(0.08, 1.0)
    #   2) bicubic resize to img_size
    #   3) random horizontal flip
    # - test:
    #   1) center crop
    #   2) bicubic resize to img_size

    @tf.function
    def train_transform(img: tf.Tensor) -> tf.Tensor:
        """Returns: tf.Tensor, shape [img_size, img_size, C], type float32"""
        img = tf.image.resize_with_pad(img, img_size, img_size,
                                       method=tf.image.ResizeMethod.BICUBIC)
        img = tf.image.random_flip_left_right(img)
        img = tf.image.random_brightness(img, max_delta=0.25)
        img = tf.image.random_contrast(img, lower=0.75, upper=1.25)
        img = tf.image.random_saturation(img, lower=0.75, upper=1.25)
        return img

    @tf.function
    def test_transform(img: tf.Tensor) -> tf.Tensor:
        """Returns: tf.Tensor, shape [img_size, img_size, C], type float32"""
        img = tf.image.resize_with_pad(img, img_size, img_size,
                                       method=tf.image.ResizeMethod.BICUBIC)
        return img

    dataloaders = {}
    for split, locs in split_to_locs.items():
        is_train = (split == 'train') and augment_train
        split_df = df[df['dataset_location'].isin(locs)]

        weights = None
        if label_weighted or weight_by_detection_conf:
            # weights sums to:
            # - if weight_by_detection_conf: (# images in split - conf delta)
            # - otherwise: (# images in split)
            weights = split_df['weights'].tolist()
            if not weight_by_detection_conf:
                assert np.isclose(sum(weights), len(split_df))

        cache: bool | str = (split in cache_splits)
        if split == 'train' and 'train' in cache_splits:
            unique_filename = str(uuid.uuid4())
            os.makedirs('/mnt/tempds/', exist_ok=True)
            cache = f'/mnt/tempds/{unique_filename}'

        ds = create_dataset(
            img_files=split_df['path'].tolist(),
            labels=split_df['label_index'].tolist(),
            sample_weights=weights,
            img_base_dir=cropped_images_dir,
            transform=train_transform if is_train else test_transform,
            target_transform=None,
            cache=cache)
        if is_train:
            ds = ds.shuffle(1000, reshuffle_each_iteration=True)
        ds = ds.batch(batch_size).prefetch(buffer_size=AUTOTUNE)
        dataloaders[split] = ds

    return dataloaders, label_names


def build_model(model_name: str, num_classes: int, img_size: int,
                pretrained: bool, finetune: bool) -> tf.keras.Model:
    """
    Creates a model with an EfficientNet base.
    """
    
    class_name = EFFICIENTNET_MODELS[model_name]['cls']
    dropout = EFFICIENTNET_MODELS[model_name]['dropout']

    model_class = tf.keras.applications.__dict__[class_name]
    weights = 'imagenet' if pretrained else None
    inputs = tf.keras.layers.Input(shape=(img_size, img_size, 3))
    base_model = model_class(
        input_tensor=inputs, weights=weights, include_top=False, pooling='avg')

    if finetune:
        # freeze the base model's weights, including BatchNorm statistics
        # https://www.tensorflow.org/guide/keras/transfer_learning#fine-tuning
        base_model.trainable = False

    # rebuild output
    x = tf.keras.layers.Dropout(dropout, name='top_dropout')(base_model.output)
    outputs = tf.keras.layers.Dense(
        num_classes,
        kernel_initializer=tf.keras.initializers.VarianceScaling(
            scale=1. / 3., mode='fan_out', distribution='uniform'),
        name='logits')(x)
    model = tf.keras.Model(inputs, outputs, name='complete_model')
    model.base_model = base_model  # cache this so that we can turn off finetune
    return model


def log_images_with_confidence(
        heap_dict: Mapping[int, list[HeapItem]],
        label_names: Sequence[str],
        epoch: int,
        tag: str) -> None:
    """
    Args:
        heap_dict: dict, maps label_id to list of HeapItem, where each HeapItem
            data is a list [img, target, top3_conf, top3_preds, img_file],
            and img is a tf.Tensor of shape [H, W, 3]
        label_names: list of str, label names in order of label id
        epoch: int
        tag: str
    """
    
    for label_id, heap in heap_dict.items():
        label_name = label_names[label_id]

        sorted_heap = sorted(heap, reverse=True)  # sort largest to smallest
        imgs_list = [item.data for item in sorted_heap]
        fig, img_files = imgs_with_confidences(imgs_list, label_names)

        # tf.summary.image requires input of shape [N, H, W, C]
        fig_img = tf.convert_to_tensor(fig_to_img(fig)[np.newaxis, ...])
        tf.summary.image(f'{label_name}/{tag}', fig_img, step=epoch)
        tf.summary.text(f'{label_name}/{tag}_files', '\n\n'.join(img_files),
                        step=epoch)


def track_extreme_examples(tp_heaps: dict[int, list[HeapItem]],
                           fp_heaps: dict[int, list[HeapItem]],
                           fn_heaps: dict[int, list[HeapItem]],
                           inputs: tf.Tensor,
                           labels: tf.Tensor,
                           img_files: tf.Tensor,
                           logits: tf.Tensor) -> None:
    """
    Updates the 5 most extreme true-positive (tp), false-positive (fp), and
    false-negative (fn) examples with examples from this batch.

    Each HeapItem's data attribute is a tuple with:
    - img: np.ndarray, shape [H, W, 3], type uint8
    - label: int
    - top3_conf: list of float
    - top3_preds: list of float
    - img_file: str

    Args:
        *_heaps: dict, maps label_id (int) to heap of HeapItems
        inputs: tf.Tensor, shape [batch_size, H, W, 3], type float32
        labels: tf.Tensor, shape [batch_size]
        img_files: tf.Tensor, shape [batch_size], type tf.string
        logits: tf.Tensor, shape [batch_size, num_classes]
    """
    
    labels = labels.numpy().tolist()
    inputs = inputs.numpy().astype(np.uint8)
    img_files = img_files.numpy().astype(str).tolist()
    batch_probs = tf.nn.softmax(logits, axis=1)
    iterable = zip(labels, inputs, img_files, batch_probs)
    for label, img, img_file, confs in iterable:
        label_conf = confs[label].numpy().item()

        top3_conf, top3_preds = tf.math.top_k(confs, k=3, sorted=True)
        top3_conf = top3_conf.numpy().tolist()
        top3_preds = top3_preds.numpy().tolist()

        data = (img, label, top3_conf, top3_preds, img_file)
        if top3_preds[0] == label:  # true positive
            item = HeapItem(priority=label_conf - top3_conf[1], data=data)
            add_to_heap(tp_heaps[label], item, k=5)
        else:
            # false positive for top3_pred[0]
            # false negative for label
            item = HeapItem(priority=top3_conf[0] - label_conf, data=data)
            add_to_heap(fp_heaps[top3_preds[0]], item, k=5)
            add_to_heap(fn_heaps[label], item, k=5)


def run_epoch(model: tf.keras.Model,
              loader: tf.data.Dataset,
              weighted: bool,
              top: Sequence[int] = (1, 3),
              loss_fn: Optional[tf.keras.losses.Loss] = None,
              weight_decay: float = 0,
              finetune: bool = False,
              optimizer: Optional[tf.keras.optimizers.Optimizer] = None,
              return_extreme_images: bool = False
              ) -> tuple[
                  dict[str, float],
                  dict[str, dict[int, list[HeapItem]]],
                  np.ndarray
              ]:
    """
    Runs for 1 epoch.

    Args:
        model: tf.keras.Model
        loader: tf.data.Dataset
        weighted: bool, whether to use sample weights in calculating loss and
            accuracy
        top: tuple of int, list of values of k for calculating top-K accuracy
        loss_fn: optional loss function, calculates the mean loss over a batch
        weight_decay: float, L2-regularization constant
        finetune: bool, if true sets model's dropout and BN layers to eval mode
        optimizer: optional optimizer

    Returns:
        metrics: dict, metrics from epoch, contains keys:
            'loss': float, mean per-example loss over entire epoch,
                only included if loss_fn is not None
            'acc_top{k}': float, accuracy@k over the entire epoch
        heaps: dict, keys are ['tp', 'fp', 'fn'], values are heap_dicts,
            each heap_dict maps label_id (int) to a heap of <= 5 HeapItems with
            data attribute (img, target, top3_conf, top3_preds, img_file)
            - 'tp': priority is the difference between target confidence and
                2nd highest confidence
            - 'fp': priority is the difference between highest confidence and
                target confidence
            - 'fn': same as 'fp'
        confusion_matrix: np.ndarray, shape [num_classes, num_classes],
            C[i, j] = # of samples with true label i, predicted as label j
    """
    # if evaluating or finetuning, set dropout & BN layers to eval mode
    is_train = False
    train_dropout_and_bn = False

    if optimizer is not None:
        assert loss_fn is not None
        is_train = True

        if not finetune:
            train_dropout_and_bn = True
            reg_vars = [
                v for v in model.trainable_variables if 'kernel' in v.name]

    if loss_fn is not None:
        losses = tf.keras.metrics.Mean()
    accuracies_topk = {
        k: tf.keras.metrics.SparseTopKCategoricalAccuracy(k) for k in top
    }

    # for each label, track 5 most-confident and least-confident examples
    tp_heaps: dict[int, list[HeapItem]] = defaultdict(list)
    fp_heaps: dict[int, list[HeapItem]] = defaultdict(list)
    fn_heaps: dict[int, list[HeapItem]] = defaultdict(list)

    all_labels = []
    all_preds = []

    tqdm_loader = tqdm.tqdm(loader)
    for batch in tqdm_loader:
        if weighted:
            inputs, labels, img_files, weights = batch
        else:
            # even if batch contains sample weights, don't use them
            inputs, labels, img_files = batch[0:3]
            weights = None

        all_labels.append(labels.numpy())
        desc = []
        with tf.GradientTape(watch_accessed_variables=is_train) as tape:
            outputs = model(inputs, training=train_dropout_and_bn)
            if loss_fn is not None:
                loss = loss_fn(labels, outputs)
                if weights is not None:
                    loss *= weights
                # we do not track L2-regularization loss in the loss metric
                losses.update_state(loss, sample_weight=weights)
                desc.append(f'Loss {losses.result().numpy():.4f}')

            if optimizer is not None:
                loss = tf.math.reduce_mean(loss)
                if not finetune:  # only regularize layers before the final FC
                    loss += weight_decay * tf.add_n(
                        tf.nn.l2_loss(v) for v in reg_vars)

        all_preds.append(tf.math.argmax(outputs, axis=1).numpy())

        if optimizer is not None:
            gradients = tape.gradient(loss, model.trainable_variables)
            optimizer.apply_gradients(zip(gradients, model.trainable_variables))

        for k, acc in accuracies_topk.items():
            acc.update_state(labels, outputs, sample_weight=weights)
            desc.append(f'Acc@{k} {acc.result().numpy() * 100:.3f}')
        tqdm_loader.set_description(' '.join(desc))

        if return_extreme_images:
            track_extreme_examples(tp_heaps, fp_heaps, fn_heaps, inputs,
                                   labels, img_files, outputs)

    confusion_matrix = sklearn.metrics.confusion_matrix(
        y_true=np.concatenate(all_labels), y_pred=np.concatenate(all_preds))

    metrics = {}
    if loss_fn is not None:
        metrics['loss'] = losses.result().numpy().item()
    for k, acc in accuracies_topk.items():
        metrics[f'acc_top{k}'] = acc.result().numpy().item() * 100
    heaps = {'tp': tp_heaps, 'fp': fp_heaps, 'fn': fn_heaps}
    return metrics, heaps, confusion_matrix


def log_run(split: str, epoch: int, writer: tf.summary.SummaryWriter,
            label_names: Sequence[str], metrics: MutableMapping[str, float],
            heaps: Mapping[str, Mapping[int, list[HeapItem]]], cm: np.ndarray
            ) -> None:
    """
    Logs the outputs (metrics, confusion matrix, tp/fp/fn images) from a
    single epoch run to Tensorboard.

    Args:
        metrics: dict, keys already prefixed with {split}/
    """
    
    per_class_recall = recall_from_confusion_matrix(cm, label_names)
    metrics.update(prefix_all_keys(per_class_recall, f'{split}/label_recall/'))

    # log metrics
    for metric, value in metrics.items():
        tf.summary.scalar(metric, value, epoch)

    # log confusion matrix
    cm_fig = plot_utils.plot_confusion_matrix(cm, classes=label_names,
                                              normalize=True)
    cm_fig_img = tf.convert_to_tensor(fig_to_img(cm_fig)[np.newaxis, ...])
    tf.summary.image(f'confusion_matrix/{split}', cm_fig_img, step=epoch)

    # log tp/fp/fn images
    for heap_type, heap_dict in heaps.items():
        log_images_with_confidence(heap_dict, label_names, epoch=epoch,
                                   tag=f'{split}/{heap_type}')
    writer.flush()


#%% Main function

def main(dataset_dir: str,
         cropped_images_dir: str,
         multilabel: bool,
         model_name: str,
         pretrained: bool,
         finetune: int,
         label_weighted: bool,
         weight_by_detection_conf: bool | str,
         epochs: int,
         batch_size: int,
         lr: float,
         weight_decay: float,
         seed: Optional[int] = None,
         logdir: str = '',
         cache_splits: Sequence[str] = ()) -> None:
    
    # input validation
    assert os.path.exists(dataset_dir)
    assert os.path.exists(cropped_images_dir)
    if isinstance(weight_by_detection_conf, str):
        assert os.path.exists(weight_by_detection_conf)

    # set seed
    seed = np.random.randint(10_000) if seed is None else seed
    np.random.seed(seed)
    tf.random.set_seed(seed)

    # create logdir and save params
    params = dict(locals())  # make a copy
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')  # '20200722_110816'
    logdir = os.path.join(logdir, timestamp)
    os.makedirs(logdir, exist_ok=True)
    print('Created logdir:', logdir)
    with open(os.path.join(logdir, 'params.json'), 'w') as f:
        json.dump(params, f, indent=1)

    gpus = tf.config.experimental.list_physical_devices('GPU')
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)

    img_size = EFFICIENTNET_MODELS[model_name]['img_size']

    # create dataloaders and log the index_to_label mapping
    loaders, label_names = create_dataloaders(
        dataset_csv_path=os.path.join(dataset_dir, 'classification_ds.csv'),
        label_index_json_path=os.path.join(dataset_dir, 'label_index.json'),
        splits_json_path=os.path.join(dataset_dir, 'splits.json'),
        cropped_images_dir=cropped_images_dir,
        img_size=img_size,
        multilabel=multilabel,
        label_weighted=label_weighted,
        weight_by_detection_conf=weight_by_detection_conf,
        batch_size=batch_size,
        augment_train=True,
        cache_splits=cache_splits)

    writer = tf.summary.create_file_writer(logdir)
    writer.set_as_default()

    model = build_model(
        model_name, num_classes=len(label_names), img_size=img_size,
        pretrained=pretrained, finetune=finetune > 0)

    # define loss function and optimizer
    loss_fn: tf.keras.losses.Loss
    if multilabel:
        loss_fn = tf.keras.losses.BinaryCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.NONE)
    else:
        loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.NONE)

    # using EfficientNet training defaults
    # - batch norm momentum: 0.99
    # - optimizer: RMSProp, decay 0.9 and momentum 0.9
    # - epochs: 350
    # - learning rate: 0.256, decays by 0.97 every 2.4 epochs
    # - weight decay: 1e-5
    lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
        lr, decay_steps=1, decay_rate=0.97, staircase=True)
    optimizer = tf.keras.optimizers.RMSprop(
        learning_rate=lr, rho=0.9, momentum=0.9)

    best_epoch_metrics: dict[str, float] = {}
    for epoch in range(epochs):
        print(f'Epoch: {epoch}')
        optimizer.learning_rate = lr_schedule(epoch)
        tf.summary.scalar('lr', optimizer.learning_rate, epoch)

        if epoch > 0 and finetune == epoch:
            print('Turning off fine-tune!')
            model.base_model.trainable = True

        print('- train:')
        
        train_metrics, train_heaps, train_cm = run_epoch(
            model, loader=loaders['train'], weighted=label_weighted,
            loss_fn=loss_fn, weight_decay=weight_decay, optimizer=optimizer,
            finetune=finetune > epoch, return_extreme_images=True)
        train_metrics = prefix_all_keys(train_metrics, prefix='train/')
        log_run('train', epoch, writer, label_names,
                metrics=train_metrics, heaps=train_heaps, cm=train_cm)

        print('- val:')
        val_metrics, val_heaps, val_cm = run_epoch(
            model, loader=loaders['val'], weighted=label_weighted,
            loss_fn=loss_fn, return_extreme_images=True)
        val_metrics = prefix_all_keys(val_metrics, prefix='val/')
        log_run('val', epoch, writer, label_names,
                metrics=val_metrics, heaps=val_heaps, cm=val_cm)

        if val_metrics['val/acc_top1'] > best_epoch_metrics.get('val/acc_top1', 0):  # pylint: disable=line-too-long
            filename = os.path.join(logdir, f'ckpt_{epoch}.h5')
            print(f'New best model! Saving checkpoint to {filename}')
            model.save(filename)
            best_epoch_metrics.update(train_metrics)
            best_epoch_metrics.update(val_metrics)
            best_epoch_metrics['epoch'] = epoch

            print('- test:')
            test_metrics, test_heaps, test_cm = run_epoch(
                model, loader=loaders['test'], weighted=label_weighted,
                loss_fn=loss_fn, return_extreme_images=True)
            test_metrics = prefix_all_keys(test_metrics, prefix='test/')
            log_run('test', epoch, writer, label_names,
                    metrics=test_metrics, heaps=test_heaps, cm=test_cm)

        # stop training after 8 epochs without improvement
        if epoch >= best_epoch_metrics['epoch'] + 8:
            break

    hparams_dict = {
        'model_name': model_name,
        'multilabel': multilabel,
        'finetune': finetune,
        'batch_size': batch_size,
        'epochs': epochs
    }
    hp.hparams(hparams_dict)
    writer.close()


#%% Command-line driver

def _parse_args() -> argparse.Namespace:
    """Parses arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Trains classifier.')
    parser.add_argument(
        'dataset_dir',
        help='path to directory containing: 1) classification dataset CSV, '
             '2) label index JSON, 3) splits JSON')
    parser.add_argument(
        'cropped_images_dir',
        help='path to local directory where image crops are saved')
    parser.add_argument(
        '--multilabel', action='store_true',
        help='for multi-label, multi-class classification')
    parser.add_argument(
        '-m', '--model-name', default='efficientnet-b0',
        choices=list(EFFICIENTNET_MODELS.keys()),
        help='which EfficientNet model')
    parser.add_argument(
        '--pretrained', action='store_true',
        help='start with pretrained model')
    parser.add_argument(
        '--finetune', type=int, default=0,
        help='only fine tune the final fully-connected layer for the first '
             '<finetune> epochs')
    parser.add_argument(
        '--label-weighted', action='store_true',
        help='weight training samples to balance labels')
    parser.add_argument(
        '--weight-by-detection-conf', nargs='?', const=True, default=False,
        help='weight training examples by detection confidence. '
             'Optionally takes a .npz file for isotonic calibration.')
    parser.add_argument(
        '--epochs', type=int, default=0,
        help='number of epochs for training, 0 for eval-only')
    parser.add_argument(
        '--batch-size', type=int, default=256,
        help='batch size for both training and eval')
    parser.add_argument(
        '--lr', type=float, default=None,
        help='initial learning rate, defaults to (0.016 * batch_size / 256)')
    parser.add_argument(
        '--weight-decay', type=float, default=1e-5,
        help='weight decay')
    parser.add_argument(
        '--seed', type=int,
        help='random seed')
    parser.add_argument(
        '--logdir', default='.',
        help='directory where TensorBoard logs and a params file are saved')
    parser.add_argument(
        '--cache', nargs='*', choices=['train', 'val', 'test'], default=(),
        help='which splits of the dataset to cache')
    return parser.parse_args()


if __name__ == '__main__':
    args = _parse_args()
    if args.lr is None:
        args.lr = 0.016 * args.batch_size / 256  # based on TF models repo
    main(dataset_dir=args.dataset_dir,
         cropped_images_dir=args.cropped_images_dir,
         multilabel=args.multilabel,
         model_name=args.model_name,
         pretrained=args.pretrained,
         finetune=args.finetune,
         label_weighted=args.label_weighted,
         weight_by_detection_conf=args.weight_by_detection_conf,
         epochs=args.epochs,
         batch_size=args.batch_size,
         lr=args.lr,
         weight_decay=args.weight_decay,
         seed=args.seed,
         logdir=args.logdir,
         cache_splits=args.cache)



"""

 mcgill_to_json.py

 Convert the .csv file provided for the McGill test data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

import pandas as pd
import os
import glob
import json
import uuid
import time
import ntpath
import humanfriendly
import PIL
import math

baseDir = r'D:\wildlife_data\mcgill_test'
input_metadata_file = os.path.join(baseDir, 'dan_500_photos_metadata.csv')
output_file = os.path.join(baseDir, 'mcgill_test.json')
image_directory = baseDir

assert(os.path.isdir(image_directory))
assert(os.path.isfile(input_metadata_file))


#%% Read source data

input_metadata = pd.read_csv(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))


#%% Map filenames to rows, verify image existence

# Create an additional column for concatenated filenames
input_metadata['relative_path'] = ''
input_metadata['full_path'] = ''

startTime = time.time()

# Maps relative filenames to rows
filenamesToRows = {}

duplicateRows = []

# Build up a map from filenames to a list of rows, checking image existence as we go
# row = input_metadata.iloc[0]
for iFile,row in input_metadata.iterrows():

    relativePath = os.path.join(row['site'],row['date_range'],str(row['camera']),
                                str(row['folder']),row['filename'])
    fullPath = os.path.join(baseDir,relativePath)
    
    if (relativePath in filenamesToRows):
        duplicateRows.append(iFile)
        filenamesToRows[relativePath].append(iFile)
    else:
        filenamesToRows[relativePath] = [iFile]
        assert(os.path.isfile(fullPath))

    row['relative_path'] = relativePath
    row['full_path'] = fullPath
    
    input_metadata.iloc[iFile] = row
    
elapsed = time.time() - startTime
print('Finished verifying image existence in {}, found {} filenames with multiple labels'.format(
      humanfriendly.format_timespan(elapsed),len(duplicateRows)))

# I didn't expect this to be true a priori, but it appears to be true, and
# it saves us the trouble of checking consistency across multiple occurrences
# of an image.
assert(len(duplicateRows) == 0)    
    
    
#%% Check for images that aren't included in the metadata file

# Enumerate all images
imageFullPaths = glob.glob(os.path.join(image_directory,'**/*.JPG'), recursive=True)

for iImage,imagePath in enumerate(imageFullPaths):
    
    imageRelPath = ntpath.relpath(imagePath, image_directory)
    assert(imageRelPath in filenamesToRows)

print('Finished checking {} images to make sure they\'re in the metadata'.format(
        len(imageFullPaths)))


#%% Create CCT dictionaries

# Also gets image sizes, so this takes ~6 minutes
#
# Implicitly checks images for overt corruptness, i.e. by not crashing.

images = []
annotations = []
categories = []

emptyCategory = {}
emptyCategory['id'] = 0
emptyCategory['name'] = 'empty'
emptyCategory['latin'] = 'empty'
emptyCategory['count'] = 0
categories.append(emptyCategory)

# Map categories to integer IDs (that's what COCO likes)
nextCategoryID = 1
labelToCategory = {'empty':emptyCategory}

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

startTime = time.time()

# row = input_metadata.iloc[0]
for iFile,row in input_metadata.iterrows():

    relPath = row['relative_path'].replace('\\','/')
    im = {}
    # Filenames look like "290716114012001a1116.jpg"
    im['id'] = relPath.replace('/','_').replace(' ','_')
    
    im['file_name'] = relPath
    
    im['seq_id'] = -1
    im['frame_num'] = -1
    
    # In the form "001a"
    im['site']= row['site']
    
    # Can be in the form '111' or 's46'
    im['camera'] = row['camera']
    
    # In the form "7/29/2016 11:40"
    im['datetime'] = row['timestamp']
    
    otherFields = ['motion','temp_F','n_present','n_waterhole','n_contact','notes']
    
    for s in otherFields:
        im[s] = row[s]
        
    # Check image height and width
    fullPath = row['full_path']
    assert(os.path.isfile(fullPath))
    pilImage = PIL.Image.open(fullPath)
    width, height = pilImage.size
    im['width'] = width
    im['height'] = height

    images.append(im)
    
    label = row['species']
    if not isinstance(label,str):
        # NaN is the only thing we should see that's not a string        
        assert math.isnan(label)
        label = 'empty'
    else:
        label = label.lower()
    
    latin = row['binomial']
    if not isinstance(latin,str):
        # NaN is the only thing we should see that's not a string
        assert math.isnan(latin)
        latin = 'empty'
    else:
        latin = latin.lower()

    if label == 'empty':
        if latin != 'empty':
            latin = 'empty'

    if label == 'unknown':
        if latin != 'unknown':
            latin = 'unknown'
            
    if label not in labelToCategory:
        print('Adding category {} ({})'.format(label,latin))
        category = {}
        categoryID = nextCategoryID
        category['id'] = categoryID        
        nextCategoryID += 1
        category['name'] = label
        category['latin'] = latin
        category['count'] = 1
        labelToCategory[label] = category
        categories.append(category)
    else:
        category = labelToCategory[label]
        category['count'] = category['count'] + 1
        categoryID = category['id']        
        
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = categoryID
    
    annotations.append(ann)
    
# ...for each image
    
# Convert categories to a CCT-style dictionary


for category in categories:
    print('Category {}, count {}'.format(category['name'],category['count']))
    
elapsed = time.time() - startTime
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = 'McGill University'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file,'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))







########
#
# string_utils.py
#
# Miscellaneous string utilities
#
########

import re

def is_float(s):
    """ 
    Checks whether a string represents a valid float
    """
    
    try:
        _ = float(s)
    except ValueError:
        return False
    return True


def human_readable_to_bytes(size):
    """
    Given a human-readable byte string (e.g. 2G, 10GB, 30MB, 20KB),
    return the number of bytes.  Will return 0 if the argument has
    unexpected form.
    
    https://gist.github.com/beugley/ccd69945346759eb6142272a6d69b4e0
    """
    
    size = re.sub(r'\s+', '', size)
    
    if (size[-1] == 'B'):
        size = size[:-1]
        
    if (size.isdigit()):
        bytes = int(size)
    elif (is_float(size)):
        bytes = float(size)
    else:
        bytes = size[:-1]
        unit = size[-1]
        try:        
            bytes = float(bytes)
            if (unit == 'T'):
                bytes *= 1024*1024*1024*1024
            elif (unit == 'G'):
                bytes *= 1024*1024*1024
            elif (unit == 'M'):
                bytes *= 1024*1024
            elif (unit == 'K'):
                bytes *= 1024
            else:
                bytes = 0
        except ValueError:
            bytes = 0
            
    return bytes



"""

add_max_conf.py

The MD output format included a "max_detection_conf" field with each image
up to and including version 1.2; it was removed as of version 1.3 (it's
redundant with the individual detection confidence values).

Just in case someone took a dependency on that field, this script allows you
to add it back to an existing .json file.

"""

#%% Imports and constants

import os
import json
from md_utils import ct_utils


#%% Main function

def add_max_conf(input_file,output_file):
    
    assert os.path.isfile(input_file), "Can't find input file {}".format(input_file)
    
    with open(input_file,'r') as f:
        d = json.load(f)
        
    for im in d['images']:
        
        max_conf = ct_utils.get_max_conf(im)
        
        if 'max_detection_conf' in im:
            assert abs(max_conf - im['max_detection_conf']) < 0.00001
        else:
            im['max_detection_conf'] = max_conf
            
    with open(output_file,'w') as f:
        json.dump(d,f,indent=1)
    

#%% Driver

import sys,argparse

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file',type=str,
                        help='Input .json file')
    parser.add_argument('output_file',type=str,
                        help='Output .json file')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    add_max_conf(args.input_file, args.output_file)

if __name__ == '__main__':
    main()




"""

 Import a Snapshot Safari project (one project, one season)

 Before running this script:

 * Mount the blob container where the images live, or copy the 
   images to local storage

 What this script does:

 * Creates a .json file
 * Creates zip archives of the season without humans.
 * Copies animals and humans to separate folders 

 After running this script:

 * Create or update LILA page
 * Push zipfile and unzipped images to LILA
 * Push unzipped humans to wildlifeblobssc
 * Delete images from UMN upload storage

 Snapshot Serengeti is handled specially, because we're dealing with bounding
 boxes too.  See snapshot_serengeti_lila.py.

"""

#%% Imports

import pandas as pd
import json
import os
import uuid
import humanfriendly
import time
import pprint
import numpy as np
import shutil

from PIL import Image
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from zipfile import ZipFile
import zipfile

from md_utils import path_utils
from md_visualization import visualize_db


#%% Constants

# project_name = 'XXX'; season_name = 'S1'; project_friendly_name = 'Snapshot Unknown'
# project_name = 'SER'; season_name = 'S1-11'; project_friendly_name = 'Snapshot Serengeti'
# project_name = 'KRU'; season_name = 'S1'; project_friendly_name = 'Snapshot Kruger'
# project_name = 'CDB'; season_name = 'S1'; project_friendly_name = 'Snapshot Camdeboo'
# project_name = 'MTZ'; season_name = 'S1'; project_friendly_name = 'Snapshot Mountain Zebra'
# project_name = 'ENO'; season_name = 'S1'; project_friendly_name = 'Snapshot Enonkishu'
# project_name = 'KAR'; season_name = 'S1'; project_friendly_name = 'Snapshot Karoo'
# project_name = 'KGA'; season_name = 'S1'; project_friendly_name = 'Snapshot Kgalagadi'
project_name = 'SER'; season_name = 'S1'; project_friendly_name = 'APN'

json_version = '2.1'

snapshot_safari_input_base = 'f:\\'
snapshot_safari_output_base = r'g:\temp\snapshot-safari-out'

category_mappings = {'blank':'empty'}

process_images_n_threads = 20

max_files_per_archive = None

#%% Folder/file creation

# E.g. KRU_S1
project_season_name = project_name + '_' + season_name

# E.g. Z:\KRU
project_base = os.path.join(snapshot_safari_input_base,project_name)
assert(os.path.isdir(project_base))

# E.g. Z:\KRU\KRU_S1
season_base = os.path.join(project_base,project_season_name)
assert(os.path.isdir(season_base))

# Contains annotations for each capture event (sequence)
annotation_file = os.path.join(project_base,project_season_name + '_report_lila.csv')

# Maps image IDs to filenames; each line looks like:
#
# KRU_S1#1#1#2,3,KRU_S1/1/1_R1/KRU_S1_1_R1_IMAG0004.JPG
image_inventory_file = os.path.join(project_base,project_season_name + '_report_lila_image_inventory.csv')

# Total number of each answer to each question, e.g. total number of times each species was identified
#
# Not used here
response_overview_file = os.path.join(project_base,project_season_name + '_report_lila_overview.csv')

assert(os.path.isfile(annotation_file))
assert(os.path.isfile(image_inventory_file))
assert(os.path.isfile(response_overview_file))

# Create output folders
assert(os.path.isdir(snapshot_safari_output_base))

output_base = os.path.join(snapshot_safari_output_base,project_name)

json_filename = os.path.join(output_base,project_friendly_name.replace(' ','') + '_' + season_name \
                             + '_v' + json_version + '.json')
species_list_filename = os.path.join(output_base,project_friendly_name.replace(' ','') + '_' + season_name \
                             + '_v' + json_version + '.species_list.csv')
summary_info_filename = os.path.join(output_base,project_friendly_name.replace(' ','') + '_' + season_name \
                             + '_v' + json_version + '.summary_info.txt')

# Images will be placed in a season-specific folder inside this (the source data includes
# this in path names)
output_public_folder = os.path.join(output_base,project_name + '_public')

output_public_zipfile = os.path.join(output_base,project_season_name + '.lila.zip')
output_private_folder = os.path.join(output_base,project_season_name + '_private')
output_preview_folder = os.path.join(output_base,project_season_name + '_preview')

os.makedirs(output_base,exist_ok=True)
os.makedirs(output_public_folder,exist_ok=True)
os.makedirs(output_private_folder,exist_ok=True)
os.makedirs(output_preview_folder,exist_ok=True)


#%% Load metadata files

image_table = pd.read_csv(image_inventory_file)
annotation_table = pd.read_csv(annotation_file)

print('Finished loading {} image mappings and {} annotations'.format(len(image_table),len(annotation_table)))


#%% Convert to dictionaries (prep)

im_id_to_image = {}
images = []
seq_id_to_images = {}    
seq_id_to_annotations = {}

annotations = []
categories = []

species_to_category = {}

empty_category_id = 0
empty_category_name = 'empty'

empty_cat = {}
empty_cat['id'] = empty_category_id
empty_cat['name'] = empty_category_name
empty_cat['count'] = 0
species_to_category['empty'] = empty_cat
categories.append(empty_cat)

next_category_id = empty_category_id + 1


#%% Convert to dictionaries (loops)
    
# iterrows() is a terrible way to do this, but this is one of those days
# where I want to get this done, not get better at Python.

print('Processing image table')

start_time = time.time()

# irow = 0; row = image_table.iloc[0]
for iRow,row in tqdm(image_table.iterrows(),total=len(image_table)):
    
    # Loaded as an int64, converting to int here
    frame_num = int(row['image_rank_in_capture'])
    assert frame_num > 0    
    sequence_id = row['capture_id']
    frame_num = int(frame_num)
    filename = row['image_path_rel']
    tokens = filename.split('.')
    assert(len(tokens)==2)
    assert(tokens[1] == 'JPG')
    id = tokens[0]
    im = {}
    im['id'] = id
    im['file_name'] = filename
    im['frame_num'] = frame_num
    im['seq_id'] = sequence_id
            
    assert id not in im_id_to_image
    im_id_to_image[id] = im
    seq_id_to_images.setdefault(sequence_id,[]).append(im)

    images.append(im)
        
# ...for each row in the image table

# Make sure image IDs are what we think they are
for im in tqdm(images):
    assert im['id'] == im['file_name'].replace('.JPG','')

print('Processing annotation table')

def is_float_and_nan(x):
    return isinstance(x,float) and np.isnan(x)

n_invalid_dates = 0
    
for iRow,row in tqdm(annotation_table.iterrows(),total=len(annotation_table)):

    sequence_id = row['capture_id']
    
    species = row['question__species'].lower()
    if species in category_mappings:
        species = category_mappings[species]
        
    category = None
    
    if species not in species_to_category:
        category = {}
        category['id'] = next_category_id
        next_category_id = next_category_id + 1
        category['name'] = species
        category['count'] = 1
        categories.append(category)
        species_to_category[species] = category
    else:
        category = species_to_category[species]
        category['count'] += 1
        
    ann = {}
    ann['sequence_level_annotation'] = True
    ann['id'] = str(uuid.uuid1())    
    ann['category_id'] = category['id']
    ann['seq_id'] = sequence_id
    
    ann['season'] = row['season']
    ann['site'] = row['site']
    if is_float_and_nan(row['capture_date_local']) or is_float_and_nan(row['capture_time_local']):
        ann['datetime'] = ''
        n_invalid_dates += 1
    else:
        ann['datetime'] = row['capture_date_local'] + ' ' + row['capture_time_local']
    ann['subject_id'] = row['subject_id']
    ann['count'] = row['question__count_median']
    ann['standing'] = row['question__standing']    
    ann['resting'] = row['question__resting']
    ann['moving'] = row['question__moving']
    ann['interacting'] = row['question__interacting']
    ann['young_present'] = row['question__young_present']    
    
    seq_id_to_annotations.setdefault(sequence_id,[]).append(ann)
    
    annotations.append(ann)

# ...for each row in the annotation table

elapsed = time.time() - start_time
print('Done converting tables to dictionaries in {}'.format(humanfriendly.format_timespan(elapsed)))

print('Converted {} annotations, {} images, {} categories ({} invalid dates)'.format(
    len(annotations),len(images),len(categories),n_invalid_dates))

    
#%% Take a look at categories

assert(len(im_id_to_image)==len(images)) 
print('Loaded metadata about {} images and {} sequences'.format(len(images),len(seq_id_to_annotations)))

categories_by_species = sorted(categories, key = lambda i: i['name'])
categories_by_count = sorted(categories, key = lambda i: i['count'])

pp = pprint.PrettyPrinter(depth=6)

# print('\nCategories by species:')
# pp.pprint(categories_by_species)
print('\nCategories by count:')
pp.pprint(categories_by_count)


#%% Fill in some image fields we didn't have when we created the image table

# width, height, corrupt, seq_num_frames, location, datetime

def process_image(im):
    
    im['width'] = -1
    im['height'] = -1    
    im['corrupt'] = False
    im['location'] = 'unknown'
    im['seq_num_frames'] = -1
    im['datetime'] = 'unknown'
    im['status'] = ''
    
    if im['seq_id'] not in seq_id_to_annotations:
        im['status'] = 'no_annotation'
        return im
    
    seq_annotations = seq_id_to_annotations[im['seq_id']]
    
    # Every annotation in this list should have the same sequence ID
    assert all(ann['seq_id'] == im['seq_id'] for ann in seq_annotations) , 'Error on image {}'.format(im['id'])

    # Figure out "seq_num_frames", which really should be done in a separate lopp;
    # there's no reason to do this redundantly for every image
    images_in_sequence = seq_id_to_images[im['seq_id']]
    
    # Every image in this sequence should point back to the same equence
    assert all(seqim['seq_id'] == im['seq_id'] for seqim in images_in_sequence), 'Error on image {}'.format(im['id'])
    
    frame_nums = [seqim['frame_num'] for seqim in images_in_sequence]
    seq_num_frames = max(frame_nums)
    im['seq_num_frames'] = seq_num_frames
    
    im['location'] = str(seq_annotations[0]['site'])
    
    # Every annotation in this list should have the same location
    assert all(str(ann['site']) == im['location'] for ann in seq_annotations), 'Error on image {}'.format(im['id'])
    
    im['datetime'] = seq_annotations[0]['datetime']
    
    # Every annotation in this list should have the same datetime
    assert all(ann['datetime'] == im['datetime'] for ann in seq_annotations), 'Error on image {}'.format(im['id'])
    
    # Is this image on disk?
    fullpath = os.path.join(project_base,im['file_name'])
    if not os.path.isfile(fullpath):
        im['status'] = 'not_on_disk'
        return im

    try:
        
        pil_im = Image.open(fullpath)
        im['height'] = pil_im.height
        im['width'] = pil_im.width
            
    except:
            
        im['corrupt'] = True        

    return im

    
if process_images_n_threads <= 1:
    
    # iImage = 0; im = images[0]    
    for iImage,im in tqdm(enumerate(images),total=len(images)):
        process_image(im)
    # ...for each image    

else:
    
    pool = ThreadPool(process_images_n_threads)
    
    # images_processed = pool.map(process_image, images)
    # images_processed = list(tqdm(pool.imap_unordered(process_image, images), total=len(images)))
    images_processed = list(tqdm(pool.imap(process_image, images), total=len(images)))
    
print('Finished adding missing fields to {} images'.format(len(images_processed)))


#%% Count missing/corrupted images    

n_missing = 0
n_corrupt = 0
n_no_annotation = 0

corrupted_images = []
missing_images = []
no_annotation_images = []

for im in tqdm(images):
    
    if im['corrupt']:
        n_corrupt += 1
        corrupted_images.append(im['file_name'])
    if im['status'] == '':
        continue
    elif im['status'] == 'not_on_disk':
        n_missing += 1
        missing_images.append(im['file_name'])
    elif im['status'] == 'no_annotation':
        n_no_annotation += 1
        no_annotation_images.append(im['file_name'])
    else:
        raise ValueError('Unrecognized status {}'.format(im['status']))

print('\nOf {} images: {} missing, {} corrupt, {} no annotation'.format(len(images),
      n_missing, n_corrupt, n_no_annotation))
    
    
#%% Print distribution of sequence lengths

seq_id_to_sequence_length = {}

for im in tqdm(images):
    
    seq_id = im['seq_id']
    seq_num_frames = im['seq_num_frames']
    if seq_id not in seq_id_to_sequence_length:
        seq_id_to_sequence_length[seq_id] = seq_num_frames

sequence_lengths = list(seq_id_to_sequence_length.values())        
    
print('\nMean/min/max sequence length is {}/{}/{}'.format(np.mean(sequence_lengths),min(sequence_lengths),max(sequence_lengths)))


#%% Replicate annotations across images

annotations_replicated = []

# iAnn = 0; ann = annotations[iAnn]
for iAnn,ann in tqdm(enumerate(annotations), total=len(annotations)):
    
    associated_images = seq_id_to_images[ann['seq_id']]
    assert len(associated_images) > 0
    for associated_image in associated_images:
        new_ann = ann.copy()
        new_ann['image_id'] = associated_image['id']
        new_ann['id'] = str(uuid.uuid1())
        annotations_replicated.append(new_ann)
        
print('\nCreated {} replicated annotations from {} original annotations'.format(len(annotations_replicated),
      len(annotations)))

annotations = annotations_replicated


#%% See what files are on disk but not annotated
    
print('Listing images from disk...')
start_time = time.time()
image_files = path_utils.find_images(project_base,bRecursive=True)
elapsed = time.time() - start_time
print('Finished listing {} files in {}'.format(len(image_files),humanfriendly.format_timespan(elapsed)))

files_not_in_db = []

for fn in tqdm(image_files):
    id = os.path.relpath(fn,project_base).replace('\\','/').replace('.JPG','')
    if id not in im_id_to_image:
        files_not_in_db.append(fn)

print('{} files not in the database (of {})'.format(len(files_not_in_db),len(image_files)))
del fn


#%% Validate image and annotation uniqueness
             
tmp_img_ids = set()
tmp_ann_ids = set()

for im in tqdm(images):
    assert im['id'] not in tmp_img_ids
    tmp_img_ids.add(im['id'])

for ann in tqdm(annotations):
    assert ann['id'] not in tmp_ann_ids
    tmp_ann_ids.add(ann['id'])
    
print('Finished uniqueness check')


#%% Minor updates to fields
    
for ann in tqdm(annotations):
    ann['location'] = ann['site']
    del ann['site']
    try:
        icount = ann['count']
    except:
        icount = -1
    ann['count'] = icount
    
for im in tqdm(images):
    del im['status']
    
for c in categories:
    del c['count']


#%% Write .json file

info = {}
info['version'] = json_version
info['description'] = 'Camera trap data from the {} program'.format(project_friendly_name)
info['date_created'] = '2019'
info['contributor'] = 'Snapshot Safari'

data = {}
data['info'] = info
data['categories'] = categories
data['annotations'] = annotations
data['images'] = images

print('Writing data to {}'.format(json_filename))

s = json.dumps(data,indent=1)
with open(json_filename, "w+") as f:
    f.write(s)

    
#%% Create a list of human files

human_image_ids = set()
human_id = species_to_category['human']['id']

# ann = annotations[0]
for ann in tqdm(annotations):
    if ann['category_id'] == human_id:
        human_image_ids.add(ann['image_id'])

print('Found {} images with humans'.format(len(human_image_ids)))


#%% Create public archive and public/private folders

debug_max_files = -1
n_dot = 1000
n_print = 10000

n_images_added = 0
zipfilename = output_public_zipfile
zip = ZipFile(zipfilename,'w')
    
print('Creating archive {}'.format(output_public_zipfile))

# im = images[0]
for iImage,im in tqdm(enumerate(images),total=len(images)):

    # E.g. KRU_S1/1/1_R1/KRU_S1_1_R1_IMAG0001.JPG
    im_relative_path = im['file_name']
    im_absolute_path = os.path.join(project_base,im_relative_path)
    assert(os.path.isfile(im_absolute_path))
    
    image_is_private = (im['id'] in human_image_ids)

    if image_is_private:
        
        # Copy to private output folder            
        output_file = os.path.join(output_private_folder,im_relative_path)
        os.makedirs(os.path.dirname(output_file),exist_ok=True)
        shutil.copyfile(src=im_absolute_path,dst=output_file)
        continue     
    
    # Add to zipfile        
    n_images_added += 1
    
    # Possibly start a new archive
    if (max_files_per_archive is not None) and (n_images_added >= max_files_per_archive):
        zip.close()
        zipfilename = zipfilename.replace('.zip','.{}.zip'.format(n_images_added))
        print('Starting new archive: {}'.format(zipfilename))
        zip = ZipFile(zipfilename,'w')
        n_images_added = 0
        
    if (n_images_added % n_dot)==0:
        print('.',end='')
    if (n_images_added % n_print)==0:
        print('{} images added to {}'.format(n_images_added,zipfilename))            
    if debug_max_files > 0 and n_images_added > debug_max_files:
        break
    
    source_file = os.path.join(project_base,im_relative_path)
    dest_file = im['file_name']
    zip.write(source_file,dest_file,zipfile.ZIP_STORED)

    # Copy to public output folder
    output_file = os.path.join(output_public_folder,im_relative_path)
    os.makedirs(os.path.dirname(output_file),exist_ok=True)        
    shutil.copyfile(src=im_absolute_path,dst=output_file)

# ...for each image

zip.close()

print('\nFinished writing {}, added {} files'.format(zipfilename,n_images_added))


#%% Consistency-check-check .json file

from data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = output_public_folder
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = False

sortedCategories, data, errorInfo = integrity_check_json_db.integrity_check_json_db(json_filename,options)

# This will produce some validation errors, because this zipfile doesn't include humans
assert(len(errorInfo['validationErrors']) == len(human_image_ids))


#%% Zip up .json and .csv files

def zip_single_file(fn,zipfilename=None):
    '''
    Zips a single file fn, by default to fn.zip
    
    Discards path information, only uses fn's base name.
    '''
    if zipfilename is None:
        zipfilename = fn + '.zip'
        
    print('Zipping {} to {}'.format(fn,zipfilename))
    with ZipFile(zipfilename,'w') as zip:
        source_file = fn
        dest_file = os.path.basename(fn)
        zip.write(source_file,dest_file,zipfile.ZIP_DEFLATED)
    return zipfilename

def zip_files_to_flat(filenames,zipfilename):
    '''
    Remove path information from everything in [filenames] and
    zip to a flat file.  Does not check uniqueness.
    '''
    with ZipFile(zipfilename,'w') as zip:
        for fn in filenames:
            source_file = fn
            dest_file = os.path.basename(fn)
            zip.write(source_file,dest_file,zipfile.ZIP_DEFLATED)

zip_single_file(json_filename)
zip_files_to_flat([annotation_file,image_inventory_file],os.path.join(output_base,project_season_name + '.csv.zip'))

print('Finished zipping .csv and .json files')


#%% When I skip to this part (using a pre-rendered .json file)

if False:

    #%%

    species_to_category = {}
    for cat in categories:
        species_to_category[cat['name']] = cat
        
    #%%
    
    human_image_ids = set()
    human_id = species_to_category['human']['id']
    
    # ann = annotations[0]
    for ann in tqdm(annotations):
        if ann['category_id'] == human_id:
            human_image_ids.add(ann['image_id'])
    
    print('Found {} images with humans'.format(len(human_image_ids)))

 
#%% Summary prep for LILA

with open(json_filename,'r') as f:
    data = json.load(f)

categories = data['categories']
annotations = data['annotations']
images = data['images']

n_empty = 0
n_species = len(categories)
n_images = len(images)

sequences = set()
for im in tqdm(images):
    sequences.add(im['seq_id'])

category_id_to_count = {}
for ann in tqdm(annotations):    
    if ann['category_id'] == 0:
        n_empty += 1
    if ann['category_id'] in category_id_to_count:
        category_id_to_count[ann['category_id']] += 1
    else:
        category_id_to_count[ann['category_id']] = 1
    
empty_categories = []    
for c in categories:
    if c['id'] in category_id_to_count:
        c['count'] = category_id_to_count[c['id']]
    else:
        empty_categories.append(c)
        c['count'] = 0

categories = [c for c in categories if c['count'] > 0]
sorted_categories = sorted(categories, key=lambda k: k['count'], reverse=True) 

with open(species_list_filename,'w') as f:
    for c in sorted_categories:
        f.write(c['name'] + ',' + str(c['count']) + '\n')

n_images = len(images) - len(human_image_ids)
n_sequences = len(sequences)
percent_empty = (100*n_empty)/len(images)
n_categories = len(categories)
top_categories = []

for i_category in range(0,len(sorted_categories)):
    c = sorted_categories[i_category]
    cat_name = c['name']
    if cat_name != 'human' and cat_name != 'empty':        
        top_categories.append(cat_name)
        if len(top_categories) == 3:
            break
    
s = 'This data set contains {} sequences of camera trap images, totaling {} images, from the {} project. Labels are provided for {} categories, primarily at the species level (for example, the most common labels are {}, {}, and {}). Approximately {:.2f}% of images are labeled as empty. A full list of species and associated image counts is available <a href="{}">here</a>.'.format(
    n_sequences,n_images,project_friendly_name,n_categories,
    top_categories[0],top_categories[1],top_categories[2],
    percent_empty,
    'https://lilablobssc.blob.core.windows.net/snapshot-safari/{}/{}_{}_v{}.species_list.csv'.format(
        project_name,project_friendly_name.replace(' ',''),season_name,json_version))
print(s)

with open(summary_info_filename,'w') as f:
    f.write(s)
    

#%% Generate preview, integrity-check labels
    
viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 5000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['test','empty']
# viz_options.classes_to_include = ['jackalblackbacked','bustardkori']
html_output_file, image_db = visualize_db.visualize_db(db_path=json_filename,
                                                         output_dir=output_preview_folder,
                                                         image_base_dir=output_public_folder,
                                                         options=viz_options)
os.startfile(html_output_file)


#%% Scrap

if False:
    
    pass

    #%% Find annotations for a particular image

    fn = missing_images[1000]
    id = fn.replace('.JPG','')
    im = im_id_to_image[id]
    seq_id = im['seq_id']
    matching_annotations = [ann for ann in annotations if ann['seq_id'] == seq_id]
    print(matching_annotations)
    
    #%% Write a list of missing images
    
    with open(os.path.join(output_base,project_name + '_' + season_name + '_missing_images.txt'), 'w') as f:
        for fn in missing_images:
            f.write('{}\n'.format(fn))
                    



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
A module to hold the configurations specific to an instance of the API.
"""

import os


#%% instance-specific API settings
# you likely need to modify these when deploying a new instance of the API

API_INSTANCE_NAME = 'cm'  # 'internal', 'cm', 'camelot', 'zooniverse'
POOL_ID = 'cm_1'  # name of the Batch pool created for this API instance

MAX_NUMBER_IMAGES_ACCEPTED_PER_JOB = 4 * 1000 * 1000  # inclusive

# Azure Batch for batch processing
BATCH_ACCOUNT_NAME = 'cameratrapssc'
BATCH_ACCOUNT_URL = 'https://cameratrapssc.southcentralus.batch.azure.com'


#%% general API settings
API_PREFIX = '/v4/camera-trap/detection-batch'  # URL to root is http://127.0.0.1:5000/v4/camera-trap/detection-batch/

MONITOR_PERIOD_MINUTES = 10

# if this number of times the thread wakes up to check is exceeded, stop the monitoring thread
MAX_MONITOR_CYCLES = 4 * 7 * int((60 * 24) / MONITOR_PERIOD_MINUTES)  # 4 weeks

IMAGE_SUFFIXES_ACCEPTED = ('.jpg', '.jpeg', '.png')  # case-insensitive
assert isinstance(IMAGE_SUFFIXES_ACCEPTED, tuple)

OUTPUT_FORMAT_VERSION = '1.1'

NUM_IMAGES_PER_TASK = 2000

OUTPUT_SAS_EXPIRATION_DAYS = 180

# quota of active Jobs in our Batch account, which all node pools i.e. API instances share;
# cannot accept job submissions if there are this many active Jobs already
MAX_BATCH_ACCOUNT_ACTIVE_JOBS = 300


#%% MegaDetector info
DETECTION_CONF_THRESHOLD = 0.1

# relative to the `megadetector_copies` folder in the container `models`
# TODO add MD versions info to AppConfig
MD_VERSIONS_TO_REL_PATH = {
    '4.1': 'megadetector_v4_1/md_v4.1.0.pb',
    '3': 'megadetector_v3/megadetector_v3_tf19.pb',
    '2': 'megadetector_v2/frozen_inference_graph.pb'
}
DEFAULT_MD_VERSION = '4.1'
assert DEFAULT_MD_VERSION in MD_VERSIONS_TO_REL_PATH

# copied from TFDetector class in detection/run_detector.py
DETECTOR_LABEL_MAP = {
    '1': 'animal',
    '2': 'person',
    '3': 'vehicle'
}


#%% Azure Batch settings
NUM_TASKS_PER_SUBMISSION = 20  # max for the Python SDK without extension is 100

NUM_TASKS_PER_RESUBMISSION = 5


#%% env variables for service credentials, and info related to these services

# Cosmos DB `batch-api-jobs` table for job status
COSMOS_ENDPOINT = os.environ['COSMOS_ENDPOINT']
COSMOS_WRITE_KEY = os.environ['COSMOS_WRITE_KEY']

# Service principal of this "application", authorized to use Azure Batch
APP_TENANT_ID = os.environ['APP_TENANT_ID']
APP_CLIENT_ID = os.environ['APP_CLIENT_ID']
APP_CLIENT_SECRET = os.environ['APP_CLIENT_SECRET']

# Blob storage account for storing Batch tasks' outputs and scoring script
STORAGE_ACCOUNT_NAME = os.environ['STORAGE_ACCOUNT_NAME']
STORAGE_ACCOUNT_KEY = os.environ['STORAGE_ACCOUNT_KEY']

# STORAGE_CONTAINER_MODELS = 'models'  # names of the two containers supporting Batch
STORAGE_CONTAINER_API = 'batch-api'

# Azure Container Registry for Docker image used by our Batch node pools
REGISTRY_SERVER = os.environ['REGISTRY_SERVER']
REGISTRY_PASSWORD = os.environ['REGISTRY_PASSWORD']
CONTAINER_IMAGE_NAME = 'cameratracrsppftkje.azurecr.io/tensorflow:1.14.0-gpu-py3'

# Azure App Configuration instance to get configurations specific to
# this instance of the API
APP_CONFIG_CONNECTION_STR = os.environ['APP_CONFIG_CONNECTION_STR']



import sys
import os

project = 'MegaDetector'
copyright = '2024, Your friendly neighborhood MegaDetector team'
author = 'Your friendly neighborhood MegaDetector team'

sys.path.insert(0, os.path.abspath("../.."))

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx_mdinclude",
    "sphinx_argparse_cli"
]

autodoc_mock_imports = ["azure", "deepdiff", "magic", "tensorflow", "pytesseract"]

myst_enable_extensions = [
    "colon_fence",
]

html_theme = 'sphinx_rtd_theme'

# collapse_navigation doesn't actually work
html_theme_options = {'navigation_depth': 2, 'collapse_navigation': False}

# html_theme = 'sphinx_book_theme'
# html_theme_options['show_navbar_depth'] = 2

# html_static_path = ['_static']

# Hide "bases: object" from all classes that don't define a base class
from sphinx.ext import autodoc

class MockedClassDocumenter(autodoc.ClassDocumenter):
    def add_line(self, line: str, source: str, *lineno: int) -> None:
        if line == "   Bases: :py:class:`object`":
            return
        super().add_line(line, source, *lineno)

autodoc.ClassDocumenter = MockedClassDocumenter


"""

 filenames_to_json.py

 Take a directory of images in which species labels are encoded by folder
 names, and produces a COCO-style .json file 
 
"""

#%% Constants and imports

import json
import io
import os
import uuid
import csv
import warnings
import datetime
from PIL import Image

from md_utils.path_utils import find_images

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings("ignore", "(Possibly )?corrupt EXIF data", UserWarning)
# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings("ignore", "Metadata warning", UserWarning)

# Filenames will be stored in the output .json relative to this base dir
baseDir = r'D:\wildlife_data\bellevue_camera_traps\bellevue_camera_traps.19.06.02.1320'
outputJsonFilename = os.path.join(baseDir,'bellevue_camera_traps.19.06.02.1320.json')
outputCsvFilename = os.path.join(baseDir,'bellevue_camera_traps.19.06.02.1320.csv')

# rawClassListFilename = os.path.join(baseDir,'bellevue_camera_traps.19.06.02.1320_classes.csv')
# classMappingsFilename = os.path.join(baseDir,'bellevue_camera_traps.19.06.02.1320_class_mapping.csv')
outputEncoding = 'utf-8'

classMappings = {'transitional':'unlabeled','moving':'unlabeled','setup':'unlabeled','blurry':'unlabeled','transitional':'unlabeled','junk':'unlabeled','unknown':'unlabeled'}

bLoadFileListIfAvailable = True

info = {}
info['year'] = 2019
info['version'] = '1.0'
info['description'] = 'Bellevue Camera Traps'
info['contributor'] = 'Dan Morris'
info['date_created'] = str(datetime.date.today())

maxFiles = -1
bReadImageSizes = False
bUseExternalRemappingTable = False


#%% Enumerate files, read image sizes

# Each element will be a list of relative path/full path/width/height
fileInfo = []
nonImages = []
nFiles = 0

if bLoadFileListIfAvailable and os.path.isfile(outputCsvFilename):
    
    print('Loading file list from {}'.format(outputCsvFilename))
    
    with open(outputCsvFilename,'r') as f:
        reader = csv.reader(f)
        csvInfo = list(list(item) for item in csv.reader(f, delimiter=','))
    
    for iRow in range(len(csvInfo)):
        csvInfo[iRow][2] = int(csvInfo[iRow][2])
        csvInfo[iRow][3] = int(csvInfo[iRow][3])
    
    fileInfo = csvInfo
    
    print('Finished reading list of {} files'.format(len(fileInfo)))
    
else:
        
    print('Enumerating files from {} to {}'.format(baseDir,outputCsvFilename))
    
    image_files = find_images(baseDir,bRecursive=True)
    print('Enumerated {} images'.format(len(image_files)))
    
    with io.open(outputCsvFilename, "w", encoding=outputEncoding) as outputFileHandle:    
    
        for fname in image_files:
      
            nFiles = nFiles + 1
            if maxFiles >= 0 and nFiles > maxFiles:            
                print('Warning: early break at {} files'.format(maxFiles))
                break
            
            fullPath = fname
            relativePath = os.path.relpath(fullPath,baseDir)
             
            if maxFiles >= 0:
                print(relativePath)
        
            h = -1
            w = -1

            if bReadImageSizes:
            
                # Read the image
                try:
                
                    im = Image.open(fullPath)
                    h = im.height
                    w = im.width
                    
                except:
                    # Corrupt or not an image
                    nonImages.append(fullPath)
                    continue
            
            # Store file info
            imageInfo = [relativePath, fullPath, w, h]
            fileInfo.append(imageInfo)
            
            # Write to output file
            outputFileHandle.write('"' + relativePath + '"' + ',' + 
                                   '"' + fullPath + '"' + ',' + 
                                   str(w) + ',' + str(h) + '\n')
                                   
        # ...for each image file
                
    # ...csv file output
        
    print("Finished writing {} file names to {}".format(nFiles,outputCsvFilename))

# ...if the file list is/isn't available
    
    
#%% Enumerate classes

# Maps classes to counts
classList = {}

for iRow,row in enumerate(fileInfo):
        
    fullPath = row[0]
    className = os.path.split(os.path.dirname(fullPath))[1]
    className = className.lower().strip()
    if className in classList:
        classList[className] += 1
    else:
        classList[className] = 1
    row.append(className)

classNames = list(classList.keys())

# We like 'empty' to be class 0
if 'empty' in classNames:
    classNames.remove('empty')    
classNames.insert(0,'empty')

print('Finished enumerating {} classes'.format(len(classList)))


#%% Assemble dictionaries

images = []
annotations = []
categories = []

categoryNameToId = {}
idToCategory = {}
imageIdToImage = {}

nextId = 0
    
for categoryName in classNames:
    
    catId = nextId
    nextId += 1
    categoryNameToId[categoryName] = catId
    newCat = {}
    newCat['id'] = categoryNameToId[categoryName]
    newCat['name'] = categoryName
    newCat['count'] = 0
    categories.append(newCat) 
    idToCategory[catId] = newCat
            
# ...for each category
    
    
# Each element is a list of relative path/full path/width/height/className
    
for iRow,row in enumerate(fileInfo):
    
    relativePath = row[0]
    w = row[2]
    h = row[3]    
    className = row[4]  
    
    assert className in categoryNameToId
    categoryId = categoryNameToId[className]
    
    im = {}
    im['id'] = str(uuid.uuid1())
    im['file_name'] = relativePath
    im['height'] = h
    im['width'] = w
    images.append(im)
    imageIdToImage[im['id']] = im
    
    ann = {}
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']
    ann['category_id'] = categoryId
    annotations.append(ann)
    
    cat = idToCategory[categoryId]
    cat['count'] += 1
    
# ...for each image

oldNameToOldId = categoryNameToId
originalCategories = categories

print('Finished assembling dictionaries')


#%% External class mapping

if bUseExternalRemappingTable:
    
    assert classMappings is None
    
    
    #%% Write raw class table
    
    # cat = categories[0]
    if os.path.isfile(rawClassListFilename):
        
        print('Not over-writing raw class table')
    
    else:
        
        with io.open(rawClassListFilename, "w", encoding=outputEncoding) as classListFileHandle:
            for cat in categories:
                catId = cat['id']
                categoryName = cat['name']
                categoryCount = cat['count']
                classListFileHandle.write(str(catId) + ',"' + categoryName + '",' + str(categoryCount) + '\n')
        
        print('Finished writing raw class table')
    
    
    #%% Read the mapped class table
    
    classMappings = {}
    
    if os.path.isfile(classMappingsFilename):
    
        print('Loading file list from {}'.format(classMappingsFilename))
            
        with open(classMappingsFilename,'r') as f:
            reader = csv.reader(f)
            mappingInfo = list(list(item) for item in csv.reader(f, delimiter=','))
        
        for mapping in mappingInfo:
            assert len(mapping) == 4
            
            # id, source, count, target
            sourceClass = mapping[1]
            targetClass = mapping[3]
            assert sourceClass not in classMappings
            classMappings[sourceClass] = targetClass
        
        print('Finished reading list of {} class mappings'.format(len(mappingInfo)))
        
else:

#%% Make classMappings contain *all* classes, not just remapped classes
    
    # cat = categories[0]
    for cat in categories:
        if cat['name'] not in classMappings:
            classMappings[cat['name']] = cat['name']

      
#%% Create new class list
    
categories = []
categoryNameToId = {}
oldIdToNewId = {}

# Start at 1, explicitly assign 0 to "empty"
nextCategoryId = 1
for sourceClass in classMappings:
    targetClass = classMappings[sourceClass]
    
    if targetClass not in categoryNameToId:
        
        if targetClass == 'empty':
            categoryId = 0
        else:
            categoryId = nextCategoryId
            nextCategoryId = nextCategoryId + 1
            
        categoryNameToId[targetClass] = categoryId
        newCat = {}
        newCat['id'] = categoryId
        newCat['name'] = targetClass
        newCat['count'] = 0
        
        if targetClass == 'empty':
            categories.insert(0,newCat)
        else:
            categories.append(newCat)

    else:
        
        categoryId = categoryNameToId[targetClass]
    
    # One-off issue with character encoding
    if sourceClass == 'humanÃ¯â‚¬Â¨':
        sourceClass = 'humanï€¨'
        
    assert sourceClass in oldNameToOldId
    oldId = oldNameToOldId[sourceClass]
    oldIdToNewId[oldId] = categoryId

categoryIdToCat = {}
for cat in categories:
    categoryIdToCat[cat['id']] = cat
    
print('Mapped {} original classes to {} new classes'.format(len(originalCategories),len(categories)))


#%% Re-map annotations
            
# ann = annotations[0]            
for ann in annotations:

    ann['category_id'] = oldIdToNewId[ann['category_id']]
    
    
#%% Write output .json

data = {}
data['info'] = info
data['images'] = images
data['annotations'] = annotations
data['categories'] = categories

json.dump(data, open(outputJsonFilename,'w'), indent=4)

print('Finished writing json to {}'.format(outputJsonFilename))


#%% Utilities

if False:
    
    #%% 
    # Find images with a particular tag
    className = 'hum'
    matches = []
    assert className in categoryNameToId
    catId = categoryNameToId[className]
    for ann in annotations:
        if ann['category_id'] == catId:
            imageId = ann['image_id']
            im = imageIdToImage[imageId]
            matches.append(im['file_name'])
    print('Found {} matches'.format(len(matches)))
    
    os.startfile(os.path.join(baseDir,matches[0]))
    
    
    #%% Randomly sample annotations
    
    import random
    nAnnotations = len(annotations)
    iAnn = random.randint(0,nAnnotations)
    ann = annotations[iAnn]
    catId = ann['category_id']
    imageId = ann['image_id']
    im = imageIdToImage[imageId]
    fn = os.path.join(baseDir,im['file_name'])
    cat = categoryIdToCat[catId]
    className = cat['name']
    print('This should be a {}'.format(className))
    os.startfile(fn)
